"""The unary dialect for Python.

This module contains the dialect for unary semantics in Python, including:

- The `UnaryOp` base class for unary operations.
- The `UAdd`, `USub`, `Not`, and `Invert` statement classes.
- The lowering pass for unary operations.
- The concrete implementation of unary operations.
- The type inference implementation of unary operations.
- The constant propagation implementation of unary operations.
- The Julia emitter for unary operations.

This dialect maps `ast.UnaryOp` nodes to the `UAdd`, `USub`, `Not`, and `Invert` statements.
"""

from . import (
    interp as interp,
    lowering as lowering,
    constprop as constprop,
    typeinfer as typeinfer,
)
from .stmts import *  # noqa: F403
from ._dialect import dialect as dialect



from kirin import ir
from kirin.passes import TypeInfer
from kirin.rewrite.abc import RewriteRule, RewriteResult

from ..stmts import Invoke, GetField
from .._dialect import dialect


@dialect.canonicalize
class ClosureField(RewriteRule):
    """Lowers captured closure fields into py.Constants.
    - Trigger on func.Invoke
    - If the callee Method has non-empty .fields, lower its func.GetField to py.Constant
    """

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not isinstance(node, Invoke):
            return RewriteResult(has_done_something=False)
        method = node.callee
        if not method.fields:
            return RewriteResult(has_done_something=False)
        # Replace func.GetField with py.Constant.
        changed = self._lower_captured_fields(method)
        if changed:
            method.fields = ()

        rewrite_result = TypeInfer(dialects=method.dialects).unsafe_run(method)
        return RewriteResult(has_done_something=changed).join(rewrite_result)

    def _get_field_index(self, getfield_stmt: GetField) -> int | None:
        fld = getfield_stmt.attributes.get("field")
        if fld:
            return getfield_stmt.field
        else:
            return None

    def _lower_captured_fields(self, method: ir.Method) -> bool:
        changed = False
        fields = method.fields
        if not fields:
            return False

        for region in method.code.regions:
            for block in region.blocks:
                for stmt in list(block.stmts):
                    if not isinstance(stmt, GetField):
                        continue
                    idx = self._get_field_index(stmt)
                    if idx is None:
                        continue
                    captured = fields[idx]
                    # Skip Methods.
                    if isinstance(captured, ir.Method):
                        continue
                    # Replace GetField with Constant.
                    from kirin.dialects import py

                    const_stmt = py.Constant(captured)
                    const_stmt.insert_before(stmt)
                    if stmt.results and const_stmt.results:
                        stmt.results[0].replace_by(const_stmt.results[0])
                    stmt.delete()
                    changed = True
        return changed



from typing import List, TypedDict
from importlib import import_module
from dataclasses import field, dataclass

from kirin import ir, types
from kirin.idtable import IdTable
from kirin.serialization.core.serializationunit import SerializationUnit

PREFIX = "_method_@"
PARAM_SEP = "->"


class MethodSymbolMeta(TypedDict, total=False):
    sym_name: str
    arg_types: List[str]
    ret_type: str


@dataclass
class SerializationContext:
    ssa_idtable: IdTable[ir.SSAValue] = field(default_factory=IdTable[ir.SSAValue])
    stmt_idtable: IdTable[ir.Statement] = field(default_factory=IdTable[ir.Statement])
    blk_idtable: IdTable[ir.Block] = field(default_factory=IdTable[ir.Block])
    region_idtable: IdTable[ir.Region] = field(default_factory=IdTable[ir.Region])

    Dialect_Lookup: dict[str, ir.Dialect] = field(default_factory=dict)
    SSA_Lookup: dict[str, ir.SSAValue] = field(default_factory=dict)
    Statement_Lookup: dict[str, ir.Statement] = field(default_factory=dict)
    Block_Lookup: dict[str, ir.Block] = field(default_factory=dict)
    Region_Lookup: dict[str, ir.Region] = field(default_factory=dict)

    Method_Symbol: dict[str, MethodSymbolMeta] = field(default_factory=dict)
    Method_Runtime: dict[str, ir.Method] = field(default_factory=dict)

    _block_reference_store: dict[str, ir.Block] = field(
        default_factory=dict[str, ir.Block]
    )

    def clear(self) -> None:
        self.SSA_Lookup.clear()
        self.Block_Lookup.clear()
        self.Region_Lookup.clear()
        self.Statement_Lookup.clear()
        self.ssa_idtable.clear()
        self.stmt_idtable.clear()
        self.blk_idtable.clear()
        self.region_idtable.clear()
        self._block_reference_store.clear()
        self.Method_Symbol.clear()
        self.Method_Runtime.clear()
        self.Dialect_Lookup.clear()


def get_str_from_type(typ: types.TypeAttribute) -> str:
    if isinstance(typ, types.PyClass):
        return typ.typ.__name__
    return "None"


def mangle(
    symbol_name: str | None,
    param_types: tuple[types.TypeAttribute, ...],
    output: types.TypeAttribute | None = None,
) -> str:
    mangled_name = f"{PREFIX}{symbol_name}"
    if param_types:
        for typ in param_types:
            mangled_name += f"{PARAM_SEP}{get_str_from_type(typ)}"
    if output is not None:
        mangled_name += f"{PARAM_SEP}{get_str_from_type(output)}"
    return mangled_name


def get_cls_from_name(serUnit: SerializationUnit) -> type:
    if not serUnit.module_name or not serUnit.class_name:
        raise ValueError(
            f"Type {serUnit.module_name} or {serUnit.class_name} cannot be None."
        )
    mod = import_module(serUnit.module_name)
    cls = getattr(mod, serUnit.class_name, None)
    if cls is None:
        if serUnit.class_name == "NoneType" and serUnit.module_name == "builtins":
            return type(None)
        else:
            raise ImportError(
                f"Could not find class {serUnit.class_name} in {serUnit.module_name}"
            )
    return cls



from typing import Callable, cast
from dataclasses import dataclass

from kirin import ir
from kirin.dialects import cf, func
from kirin.rewrite.abc import RewriteRule, RewriteResult

# TODO: use func.Constant instead of kirin.dialects.py.stmts.Constant
from kirin.dialects.py.constant import Constant

# NOTE: this only inlines func dialect


@dataclass
class Inline(RewriteRule):
    heuristic: Callable[[ir.Statement], bool]
    """inline heuristic that determines whether a function should be inlined
    """

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if isinstance(node, func.Invoke):
            return self.rewrite_func_Invoke(node)
        elif isinstance(node, func.Call):
            return self.rewrite_func_Call(node)
        else:
            return RewriteResult()

    def rewrite_func_Call(self, node: func.Call) -> RewriteResult:
        if not isinstance(lambda_stmt := node.callee.owner, func.Lambda):
            return RewriteResult()

        # NOTE: a lambda statement is defined and used in the same scope
        self.inline_call_like(node, tuple(node.args), lambda_stmt.body)
        return RewriteResult(has_done_something=True)

    def rewrite_func_Invoke(self, node: func.Invoke) -> RewriteResult:
        has_done_something = False
        callee = node.callee

        if (
            isinstance(callee, ir.Method)
            and self.heuristic(callee.code)
            and (call_trait := callee.code.get_trait(ir.CallableStmtInterface))
            is not None
        ):
            region = call_trait.get_callable_region(callee.code)
            func_self = Constant(node.callee)
            func_self.result.name = node.callee.sym_name
            func_self.insert_before(node)
            self.inline_call_like(
                node, (func_self.result,) + tuple(arg for arg in node.args), region
            )
            has_done_something = True

        return RewriteResult(has_done_something=has_done_something)

    def inline_call_like(
        self,
        call_like: ir.Statement,
        args: tuple[ir.SSAValue, ...],
        region: ir.Region,
    ):
        """
        Inline a function call-like statement

        Args:
            call_like (ir.Statement): the call-like statement
            args (tuple[ir.SSAValue, ...]): the arguments of the call (first one is the callee)
            region (ir.Region): the region of the callee
        """
        if not call_like.parent_block:
            return

        if not call_like.parent_region:
            return

        # NOTE: we cannot change region because it may be used elsewhere
        inline_region: ir.Region = region.clone()

        # Preserve source information by attributing inlined code to the call site
        if call_like.source is not None:
            for block in inline_region.blocks:
                if block.source is None:
                    block.source = call_like.source
                for stmt in block.stmts:
                    if stmt.source is None:
                        stmt.source = call_like.source

        if self._can_use_simple_inline(inline_region):
            return self._inline_simple(call_like, args, inline_region.blocks[0])

        return self._inline_complex(call_like, args, inline_region)

    def _can_use_simple_inline(self, inline_region: ir.Region) -> bool:
        """Check if we can use the fast path for simple single-block inlining.

        Args:
            inline_region: The cloned region to be inlined

        Returns:
            True if simple inline is possible (single block with simple return)
        """
        if len(inline_region.blocks) != 1:
            return False

        block = inline_region.blocks[0]

        # Last statement must be a simple return
        if not isinstance(block.last_stmt, func.Return):
            return False

        return True

    def _inline_simple(
        self,
        call_like: ir.Statement,
        args: tuple[ir.SSAValue, ...],
        func_block: ir.Block,
    ):
        """Fast path: inline single-block function by splicing statements.

        For simple functions with no control flow, we just clone the function's
        statements and insert them before the call site.
        No new blocks are created, no statement parent updates are needed.

        Complexity: O(k) where k = number of statements in function (typically small)

        Args:
            call_like: The call statement to replace
            args: Arguments to the call (first is callee, rest are parameters)
            func_block: The single block from the cloned function region
        """
        ssa_map: dict[ir.SSAValue, ir.SSAValue] = {}
        for func_arg, call_arg in zip(func_block.args, args):
            ssa_map[func_arg] = call_arg
            if func_arg.name and call_arg.name is None:
                call_arg.name = func_arg.name

        for stmt in func_block.stmts:
            if isinstance(stmt, func.Return):
                return_value = ssa_map.get(stmt.value, stmt.value)

                if call_like.results:
                    for call_result in call_like.results:
                        call_result.replace_by(return_value)

                # Don't insert the return statement itself
                break

            new_stmt = stmt.from_stmt(
                stmt,
                args=[ssa_map.get(arg, arg) for arg in stmt.args],
                regions=[r.clone(ssa_map) for r in stmt.regions],
                successors=stmt.successors,  # successors are empty for simple stmts
            )

            new_stmt.insert_before(call_like)

            # Update SSA mapping for newly created results
            for old_result, new_result in zip(stmt.results, new_stmt.results):
                ssa_map[old_result] = new_result
                if old_result.name:
                    new_result.name = old_result.name

        call_like.delete()

    def _inline_complex(
        self,
        call_like: ir.Statement,
        args: tuple[ir.SSAValue, ...],
        inline_region: ir.Region,
    ):
        """Inline multi-block function with control flow.

        This handles the general case where the function has multiple blocks

        Complexity: O(n+k) where n = statements after call site (due to moving them)
            and k = number of statements in function.

        Args:
            call_like: The call statement to replace
            args: Arguments to the call
            inline_region: The cloned function region to inline
        """

        # <stmt>
        # <stmt>
        # <br (a, b, c)>

        # <block (a, b,c)>:
        # <block>:
        # <block>:
        # <br>

        # ^<block>:
        # <stmt>
        # <stmt>

        # 1. we insert the entry block of the callee function
        # 2. we insert the rest of the blocks into the parent region
        # 3.1 if the return is in the entry block, means no control flow,
        #     replace the call results with the return values
        # 3.2 if the return is some of the blocks, means control flow,
        #     split the current block into two, and replace the return with
        #     the branch instruction
        # 4. remove the call
        parent_block: ir.Block = cast(ir.Block, call_like.parent_block)
        parent_region: ir.Region = cast(ir.Region, call_like.parent_region)

        # wrap what's after invoke into a block
        after_block = ir.Block()
        stmt = call_like.next_stmt
        while stmt is not None:
            stmt.detach()
            after_block.stmts.append(stmt)
            stmt = call_like.next_stmt

        for result in call_like.results:
            block_arg = after_block.args.append_from(result.type, result.name)
            result.replace_by(block_arg)

        parent_block_idx = parent_region._block_idx[parent_block]
        entry_block = inline_region.blocks.popfirst()
        idx, block = 0, entry_block
        while block is not None:
            idx += 1

            if block.last_stmt and isinstance(block.last_stmt, func.Return):
                block.last_stmt.replace_by(
                    cf.Branch(
                        arguments=(block.last_stmt.value,),
                        successor=after_block,
                    )
                )

            parent_region.blocks.insert(parent_block_idx + idx, block)
            block = inline_region.blocks.popfirst()

        parent_region.blocks.append(after_block)

        # NOTE: we expect always to have an entry block
        # but we still check for it cuz we are not gonna
        # error for empty regions here.
        if entry_block:
            cf.Branch(
                arguments=args,
                successor=entry_block,
            ).insert_before(call_like)
        call_like.delete()



from typing import Any

import pytest

from kirin import ir, types, lowering
from kirin.decl import info, statement
from kirin.prelude import python_no_opt
from kirin.dialects import func

T = types.TypeVar("T")

dialect = ir.Dialect("test")


@statement(dialect=dialect, repr=True)
class DummyStatement(ir.Statement):
    name = "dummy"
    traits = frozenset({ir.Pure(), ir.ConstantLike(), lowering.FromPythonCall()})

    # args
    noinfo: ir.SSAValue
    vararg_noinfo: tuple[ir.SSAValue, ...]
    xxx: ir.SSAValue = info.argument(T)
    xxx_Any: ir.SSAValue = info.argument()
    xxx_vararg: tuple[ir.SSAValue, ...] = info.argument()

    # results
    xxx_result: ir.ResultValue = info.result(T)

    # attributes
    xxx_property: Any = info.attribute(default="")
    xxx_attribute: bool = info.attribute()


def dummy(
    noinfo,
    vararg_noinfo,
    xxx,
    xxx_Any,
    xxx_vararg,
):
    return DummyStatement(
        noinfo,
        (vararg_noinfo,),
        xxx,
        xxx_Any,
        xxx_vararg=(xxx_vararg,),
        xxx_property="xxx_property",
        xxx_attribute=False,
    )


def dummy_2(
    noinfo,
    vararg_noinfo,
    xxx,
    xxx_Any,
    xxx_vararg,
):
    return DummyStatement(
        noinfo,
        (vararg_noinfo,),
        xxx,
        xxx_Any=xxx_Any,
        xxx_vararg=(xxx_vararg,),
        xxx_property="xxx_property",
        xxx_attribute=False,
    )


def non_const(
    noinfo,
    vararg_noinfo,
    xxx,
    xxx_Any,
    xxx_vararg,
    non_const,
):
    return DummyStatement(
        noinfo,
        (vararg_noinfo,),
        xxx,
        xxx_Any=xxx_Any,
        xxx_vararg=(xxx_vararg,),
        xxx_property=non_const,
        xxx_attribute=False,
    )


def no_group(
    noinfo,
    vararg_noinfo,
    xxx,
    xxx_Any,
    xxx_vararg,
):
    return DummyStatement(
        noinfo,
        vararg_noinfo,
        xxx,
        xxx_Any=xxx_Any,
        xxx_vararg=(xxx_vararg,),
        xxx_property="non_const",
        xxx_attribute=False,
    )


def test_from_python_call():
    assert DummyStatement.dialect is dialect
    lower = lowering.Python(python_no_opt.data.union([func, dialect]))

    func_ = dummy
    code = lower.python_function(func_)
    assert isinstance(code, func.Function)
    block = code.body.blocks[0]
    stmt: DummyStatement = block.stmts.at(0)  # type: ignore
    assert stmt.noinfo is block.args[1]
    assert stmt.vararg_noinfo == (block.args[2],)
    assert stmt.xxx is block.args[3]
    assert stmt.xxx_Any is block.args[4]
    assert stmt.xxx_vararg == (block.args[5],)
    assert stmt.xxx_result.type == T
    assert stmt.xxx_property == "xxx_property"
    assert stmt.xxx_attribute is False  # type: ignore

    code = lower.python_function(dummy_2)
    assert isinstance(code, func.Function)
    block = code.body.blocks[0]
    stmt: DummyStatement = block.stmts.at(0)  # type: ignore
    assert stmt.noinfo is block.args[1]
    assert stmt.vararg_noinfo == (block.args[2],)
    assert stmt.xxx is block.args[3]
    assert stmt.xxx_Any is block.args[4]
    assert stmt.xxx_vararg == (block.args[5],)
    assert stmt.xxx_result.type == T
    assert stmt.xxx_property == "xxx_property"
    assert stmt.xxx_attribute is False  # type: ignore

    with pytest.raises(lowering.BuildError):
        lower.python_function(non_const)

    with pytest.raises(lowering.BuildError):
        lower.python_function(no_group)



class SerializationUnit:
    kind: str
    module_name: str
    class_name: str
    data: dict

    def __init__(self, kind: str, module_name: str, class_name: str, data: dict):
        self.kind = kind
        self.module_name = module_name
        self.class_name = class_name
        self.data = data



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Iterable, Iterator, cast
from dataclasses import field, dataclass

from typing_extensions import Self

from kirin.ir.ssa import SSAValue
from kirin.source import SourceInfo
from kirin.ir.exception import ValidationError
from kirin.ir.nodes.base import IRNode
from kirin.ir.nodes.view import MutableSequenceView
from kirin.ir.nodes.block import Block

if TYPE_CHECKING:
    from kirin.print import Printer
    from kirin.ir.nodes.stmt import Statement
    from kirin.serialization.base.serializer import Serializer
    from kirin.serialization.base.deserializer import Deserializer
    from kirin.serialization.core.serializationunit import SerializationUnit


@dataclass
class RegionBlocks(MutableSequenceView[list[Block], "Region", Block]):
    """A View object that contains a list of Blocks of a Region.

    Description:
        This is a proxy object that provide safe API to manipulate the Blocks of a Region.

    """

    def __setitem__(
        self, idx: int | slice, block_or_blocks: Block | Iterable[Block]
    ) -> None:
        """Replace/Set the Blocks of the Region.

        Args:
            idx (int | slice): The index or slice to replace the [`Blocks`][kirin.ir.Block].
            block_or_blocks (Block | Iterable[Block]): The Block or Blocks to replace the Blocks.

        """
        if isinstance(idx, int) and isinstance(block_or_blocks, Block):
            self.field[idx].detach()
            block_or_blocks.attach(self.node)
            self.field[idx] = block_or_blocks
            self.node._block_idx[block_or_blocks] = idx
        elif isinstance(idx, slice) and isinstance(block_or_blocks, Iterable):
            for block in block_or_blocks:
                block.attach(self.node)
            self.field[idx] = block_or_blocks
            self.node._block_idx = {
                block: i for i, block in enumerate(self.field)
            }  # reindex
        else:
            raise ValueError("Invalid assignment")

    def __delitem__(self, idx: int) -> None:
        self.field[idx].delete()

    def pop(self, idx: int = -1) -> Block:
        item = self.field[idx]
        self[idx].detach()
        return item

    def insert(self, idx: int, value: Block) -> None:
        """Inserts a Block at the specified index.

        Args:
            idx (int): The index at which to insert the block.
            value (Block): The block to be inserted.
        """
        value.attach(self.node)
        self.field.insert(idx, value)
        for i, value in enumerate(self.field[idx:], idx):
            self.node._block_idx[value] = i

    def append(self, value: Block) -> None:
        """Append a Block to the Region.

        Args:
            value (Block): The block to be appended.
        """
        value.attach(self.node)
        self.node._block_idx[value] = len(self.field)
        self.field.append(value)


@dataclass
class Region(IRNode["Statement"]):
    """Region consist of a list of Blocks

    !!! note "Pretty Printing"
        This object is pretty printable via
        [`.print()`][kirin.print.printable.Printable.print] method.
    """

    IS_REGION: ClassVar[bool] = True

    _blocks: list[Block] = field(default_factory=list, repr=False)
    _block_idx: dict[Block, int] = field(default_factory=dict, repr=False)
    _parent: Statement | None = field(default=None, repr=False)

    def __init__(
        self,
        blocks: Block | Iterable[Block] = (),
        parent: Statement | None = None,
        *,
        source: SourceInfo | None = None,
    ):
        """Initialize a Region object.

        Args:
            blocks (Block | Iterable[Block], optional): A single [`Block`][kirin.ir.Block] object or an iterable of Block objects. Defaults to ().
            parent (Statement | None, optional): The parent [`Statement`][kirin.ir.Statement] object. Defaults to None.
        """
        super().__init__()
        self.source = source
        self._blocks = []
        self._block_idx = {}
        self.parent_node = parent
        if isinstance(blocks, Block):
            blocks = (blocks,)
        for block in blocks:
            self.blocks.append(block)

    def __getitem__(self, block: Block) -> int:
        """Get the index of a block within the region.

        Args:
            block (Block): The block to get the index of.

        Raises:
            ValueError: If the block does not belong to the region.

        Returns:
            int: The index of the block within the region.
        """
        if block.parent is not self:
            raise ValueError("Block does not belong to the region")
        return self._block_idx[block]

    def __hash__(self) -> int:
        return id(self)

    def clone(self, ssamap: dict[SSAValue, SSAValue] | None = None) -> Region:
        """Clone a region. This will clone all blocks and statements in the region.
        `SSAValue` defined outside the region will not be cloned unless provided in `ssamap`.
        """
        ret = Region()
        successor_map: dict[Block, Block] = {}
        _ssamap = ssamap or {}
        for block in self.blocks:
            new_block = Block()
            ret.blocks.append(new_block)
            successor_map[block] = new_block
            for arg in block.args:
                new_arg = new_block.args.append_from(arg.type, arg.name)
                _ssamap[arg] = new_arg

        # update statements
        for block in self.blocks:
            for stmt in block.stmts:
                new_stmt = stmt.from_stmt(
                    stmt,
                    args=[_ssamap.get(arg, arg) for arg in stmt.args],
                    regions=[region.clone(_ssamap) for region in stmt.regions],
                    successors=[
                        successor_map[successor] for successor in stmt.successors
                    ],
                )
                successor_map[block].stmts.append(new_stmt)
                for result, new_result in zip(stmt.results, new_stmt.results):
                    _ssamap[result] = new_result
                    new_result.name = result.name

        return ret

    @property
    def parent_node(self) -> Statement | None:
        """Get the parent statement of the region."""
        return self._parent

    @parent_node.setter
    def parent_node(self, parent: Statement | None) -> None:
        """Set the parent statement of the region."""
        from kirin.ir.nodes.stmt import Statement

        self.assert_parent(Statement, parent)
        self._parent = parent

    @property
    def blocks(self) -> RegionBlocks:
        """Get the Blocks in the region.

        Returns:
            RegionBlocks: The blocks View object of the region.
        """
        return RegionBlocks(self, self._blocks)

    @property
    def region_index(self) -> int:
        """Get the index of the region within the parent scope.

        Returns:
            int: The index of the region within the parent scope.
        """
        if self.parent_node is None:
            raise ValueError("Region has no parent")
        for idx, region in enumerate(self.parent_node.regions):
            if region is self:
                return idx
        raise ValueError("Region not found in parent")

    def detach(self, index: int | None = None) -> None:
        """Detach this Region from the IR tree graph.

        Note:
            Detach only detach the Region from the IR graph. It does not remove uses that reference the Region.
        """
        # already detached
        if self.parent_node is None:
            return

        if index is not None:
            region_idx = index
        else:
            region_idx = self.region_index

        del self.parent_node._regions[region_idx]
        self.parent_node = None

    def drop_all_references(self) -> None:
        """Remove all the dependency that reference/uses this Region."""
        self.parent_node = None
        for block in self._blocks:
            block.drop_all_references()

    def delete(self, safe: bool = True) -> None:
        """Delete the Region completely from the IR graph.

        Note:
            This method will detach + remove references of the Region.

        Args:
            safe (bool, optional): If True, raise error if there is anything that still reference components in the Region. Defaults to True.
        """
        self.detach()
        self.drop_all_references()

    def is_structurally_equal(  # pyright: ignore[reportIncompatibleMethodOverride]
        self,
        other: Self,
        context: dict[IRNode | SSAValue, IRNode | SSAValue] | None = None,
    ) -> bool:
        """Check if the Region is structurally equal to another Region.

        Args:
            other (Self): The other Region to compare with.
            context (dict[IRNode  |  SSAValue, IRNode  |  SSAValue] | None, optional): A map of IRNode/SSAValue to hint that they are equivalent so the check will treat them as equivalent. Defaults to None.

        Returns:
            bool: True if the Region is structurally equal to the other Region.
        """
        if context is None:
            context = {}
        context[self] = other
        if len(self.blocks) != len(other.blocks):
            return False

        for block, other_block in zip(self.blocks, other.blocks):
            context[block] = other_block

        if not all(
            block.is_structurally_equal(other_block, context)
            for block, other_block in zip(self.blocks, other.blocks)
        ):
            return False

        return True

    def walk(
        self, *, reverse: bool = False, region_first: bool = False
    ) -> Iterator[Statement]:
        """Traversal the Statements of Blocks in the Region.

        Args:
            reverse (bool, optional): If walk in the reversed manner. Defaults to False.
            region_first (bool, optional): If the walk should go through the Statement first or the Region of a Statement first. Defaults to False.

        Yields:
            Iterator[Statement]: An iterator that yield Statements of Blocks in the Region, in the specified order.
        """
        for block in reversed(self.blocks) if reverse else self.blocks:
            yield from block.walk(reverse=reverse, region_first=region_first)

    def stmts(self) -> Iterator[Statement]:
        """Iterate over all the Statements in the Region. This does not walk into nested Regions.

        Yields:
            Iterator[Statement]: An iterator that yield Statements of Blocks in the Region.
        """
        for block in self.blocks:
            yield from block.stmts

    def print_impl(self, printer: Printer) -> None:
        # populate block ids
        for block in self.blocks:
            printer.state.block_id[block]

        printer.plain_print("{")
        if len(self.blocks) == 0:
            printer.print_newline()
            printer.plain_print("}")
            return

        with printer.align(printer.result_width(self.stmts())):
            with printer.indent(increase=2, mark=True):
                printer.print_newline()
                for idx, bb in enumerate(self.blocks):
                    printer.print(bb)

                    if idx != len(self.blocks) - 1:
                        printer.print_newline()

        printer.print_newline()
        printer.plain_print("}")

    def verify(self) -> None:
        """Verify the correctness of the Region.

        Raises:
            ValidationError: If the Region is not correct.
        """
        from kirin.ir.nodes.stmt import Statement

        if not isinstance(self.parent_node, Statement):
            raise ValidationError(
                self, "expect Region to have a parent of type Statement"
            )

        for block in self.blocks:
            block.verify()

    def verify_type(self) -> None:
        for block in self.blocks:
            block.verify_type()

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return serializer.serialize_region(self)

    @classmethod
    def deserialize(
        cls: type[Self], serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> Self:
        return cast(Self, deserializer.deserialize_region(serUnit))



from dataclasses import dataclass

from kirin import ir
from kirin.rewrite import abc
from kirin.dialects import py

from .stmts import X, Y, Z, Id, PauliOperator


@dataclass
class RewritePauliMult(abc.RewriteRule):
    def rewrite_Statement(self, node: ir.Statement) -> abc.RewriteResult:
        if not isinstance(node, py.binop.Mult):
            return abc.RewriteResult()

        if not isinstance(node.lhs.owner, PauliOperator) and not isinstance(
            node.rhs.owner, PauliOperator
        ):
            return abc.RewriteResult()

        if isinstance(node.lhs.owner, py.Constant):
            assert isinstance(node.rhs.owner, PauliOperator)  # make the linter happy
            new_op = self.number_pauli_mult(node.lhs.owner, node.rhs.owner)
            node.replace_by(new_op)
            return abc.RewriteResult(has_done_something=True)
        elif isinstance(node.rhs.owner, py.Constant):
            assert isinstance(node.lhs.owner, PauliOperator)  # make the linter happy
            new_op = self.number_pauli_mult(node.rhs.owner, node.lhs.owner)
            node.replace_by(new_op)
            return abc.RewriteResult(has_done_something=True)

        if not isinstance(node.lhs.owner, PauliOperator) or not isinstance(
            node.rhs.owner, PauliOperator
        ):
            return abc.RewriteResult()

        new_op = self.pauli_pauli_mult(node.lhs.owner, node.rhs.owner)
        node.replace_by(new_op)
        return abc.RewriteResult(has_done_something=True)

    @staticmethod
    def number_pauli_mult(lhs: py.Constant, rhs: PauliOperator) -> PauliOperator:
        num = lhs.value.unwrap() * rhs.pre_factor
        return type(rhs)(pre_factor=num)

    @staticmethod
    def pauli_pauli_mult(lhs: PauliOperator, rhs: PauliOperator) -> PauliOperator:
        num = rhs.pre_factor * lhs.pre_factor

        if isinstance(lhs, type(rhs)):
            return Id(pre_factor=num)

        if isinstance(lhs, type(rhs)):
            return Id(pre_factor=num)

        if isinstance(lhs, Id):
            return type(rhs)(pre_factor=num)

        if isinstance(rhs, Id):
            return type(lhs)(pre_factor=num)

        if isinstance(lhs, X):
            if isinstance(rhs, Y):
                return Z(pre_factor=1j * num)
            elif isinstance(rhs, Z):
                return Y(pre_factor=-1j * num)

        if isinstance(lhs, Y):
            if isinstance(rhs, X):
                return Z(pre_factor=-1j * num)
            elif isinstance(rhs, Z):
                return X(pre_factor=1j * num)

        if isinstance(lhs, Z):
            if isinstance(rhs, Y):
                return X(pre_factor=-1j * num)
            elif isinstance(rhs, X):
                return Y(pre_factor=1j * num)

        raise RuntimeError("How on earth did we end up here?")



from kirin import ir, lowering
from kirin.decl import info, statement
from kirin.prelude import basic_no_opt
from kirin.rewrite.cse import _HASHABLE_SLICE, Info, CommonSubexpressionElimination
from kirin.rewrite.walk import Walk

dialect = ir.Dialect("test")


@statement(dialect=dialect)
class MultiResult(ir.Statement):
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    result_a: ir.ResultValue = info.result()
    result_b: ir.ResultValue = info.result()


dummy_dialect = basic_no_opt.add(dialect)


def test_multi_result():
    @dummy_dialect
    def duplicated():
        x, y = MultiResult()  # type: ignore
        a, b = MultiResult()  # type: ignore
        return x + a, y + b

    stmt_0 = duplicated.callable_region.blocks[0].stmts.at(0)
    stmt_1 = duplicated.callable_region.blocks[0].stmts.at(1)
    assert isinstance(stmt_0, MultiResult)
    assert isinstance(stmt_1, MultiResult)

    Walk(CommonSubexpressionElimination()).rewrite(duplicated.code)

    stmt_0 = duplicated.callable_region.blocks[0].stmts.at(0)
    stmt_1 = duplicated.callable_region.blocks[0].stmts.at(1)
    assert isinstance(stmt_0, MultiResult)
    assert not isinstance(stmt_1, MultiResult)


def test_info():
    info_value = Info(ir.Statement, (), (ir.PyAttr(slice(None)),), (), ())

    if not _HASHABLE_SLICE:
        assert info_value._hashable is False
        assert info_value._hash == id(info_value)
    else:
        assert info_value._hashable is True
        info_value._hash = hash((ir.Statement,) + (ir.PyAttr(slice(None)),))



import numpy as np
from scipy import special

from kirin import types
from kirin.interp import Frame, MethodTable, impl
from kirin.dialects import ilist

from . import stmts
from ._dialect import dialect


@dialect.register
class MathMethodTable(MethodTable):

    @impl(stmts.add)
    def add(self, interp, frame: Frame, stmt: stmts.add):
        lhs = frame.get(stmt.lhs)
        rhs = frame.get(stmt.rhs)
        if isinstance(lhs, ilist.IList):
            lhs = np.asarray(lhs)
        if isinstance(rhs, ilist.IList):
            rhs = np.asarray(rhs)
        result = lhs + rhs
        return (ilist.IList(result.tolist(), elem=types.Float),)

    @impl(stmts.acos)
    def acos(self, interp, frame: Frame, stmt: stmts.acos):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(np.arccos(np.asarray(values[0])).tolist(), elem=types.Float),
        )

    @impl(stmts.asin)
    def asin(self, interp, frame: Frame, stmt: stmts.asin):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(np.arcsin(np.asarray(values[0])).tolist(), elem=types.Float),
        )

    @impl(stmts.asinh)
    def asinh(self, interp, frame: Frame, stmt: stmts.asinh):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(np.arcsinh(np.asarray(values[0])).tolist(), elem=types.Float),
        )

    @impl(stmts.atan)
    def atan(self, interp, frame: Frame, stmt: stmts.atan):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(np.arctan(np.asarray(values[0])).tolist(), elem=types.Float),
        )

    @impl(stmts.atan2)
    def atan2(self, interp, frame: Frame, stmt: stmts.atan2):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(
                np.arctan2(np.asarray(values[0]), np.asarray(values[1])).tolist(),
                elem=types.Float,
            ),
        )

    @impl(stmts.atanh)
    def atanh(self, interp, frame: Frame, stmt: stmts.atanh):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(np.arctanh(np.asarray(values[0])).tolist(), elem=types.Float),
        )

    @impl(stmts.ceil)
    def ceil(self, interp, frame: Frame, stmt: stmts.ceil):
        values = frame.get_values(stmt.args)
        return (ilist.IList(np.ceil(np.asarray(values[0])).tolist(), elem=types.Float),)

    @impl(stmts.copysign)
    def copysign(self, interp, frame: Frame, stmt: stmts.copysign):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(
                np.copysign(np.asarray(values[0]), np.asarray(values[1])).tolist(),
                elem=types.Float,
            ),
        )

    @impl(stmts.cos)
    def cos(self, interp, frame: Frame, stmt: stmts.cos):
        values = frame.get_values(stmt.args)
        return (ilist.IList(np.cos(np.asarray(values[0])).tolist(), elem=types.Float),)

    @impl(stmts.cosh)
    def cosh(self, interp, frame: Frame, stmt: stmts.cosh):
        values = frame.get_values(stmt.args)
        return (ilist.IList(np.cosh(np.asarray(values[0])).tolist(), elem=types.Float),)

    @impl(stmts.degrees)
    def degrees(self, interp, frame: Frame, stmt: stmts.degrees):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(np.degrees(np.asarray(values[0])).tolist(), elem=types.Float),
        )

    @impl(stmts.div)
    def div(self, interp, frame: Frame, stmt: stmts.div):
        lhs = frame.get(stmt.lhs)
        rhs = frame.get(stmt.rhs)
        if isinstance(lhs, ilist.IList):
            lhs = np.asarray(lhs)
        if isinstance(rhs, ilist.IList):
            rhs = np.asarray(rhs)
        result = lhs / rhs
        return (ilist.IList(result.tolist(), elem=types.Float),)

    @impl(stmts.erf)
    def erf(self, interp, frame: Frame, stmt: stmts.erf):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(special.erf(np.asarray(values[0])).tolist(), elem=types.Float),
        )

    @impl(stmts.erfc)
    def erfc(self, interp, frame: Frame, stmt: stmts.erfc):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(special.erfc(np.asarray(values[0])).tolist(), elem=types.Float),
        )

    @impl(stmts.exp)
    def exp(self, interp, frame: Frame, stmt: stmts.exp):
        values = frame.get_values(stmt.args)
        return (ilist.IList(np.exp(np.asarray(values[0])).tolist(), elem=types.Float),)

    @impl(stmts.expm1)
    def expm1(self, interp, frame: Frame, stmt: stmts.expm1):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(np.expm1(np.asarray(values[0])).tolist(), elem=types.Float),
        )

    @impl(stmts.fabs)
    def fabs(self, interp, frame: Frame, stmt: stmts.fabs):
        values = frame.get_values(stmt.args)
        return (ilist.IList(np.fabs(np.asarray(values[0])).tolist(), elem=types.Float),)

    @impl(stmts.floor)
    def floor(self, interp, frame: Frame, stmt: stmts.floor):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(np.floor(np.asarray(values[0])).tolist(), elem=types.Float),
        )

    @impl(stmts.fmod)
    def fmod(self, interp, frame: Frame, stmt: stmts.fmod):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(
                np.fmod(np.asarray(values[0]), np.asarray(values[1])).tolist(),
                elem=types.Float,
            ),
        )

    @impl(stmts.gamma)
    def gamma(self, interp, frame: Frame, stmt: stmts.gamma):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(
                special.gamma(np.asarray(values[0])).tolist(), elem=types.Float
            ),
        )

    @impl(stmts.isfinite)
    def isfinite(self, interp, frame: Frame, stmt: stmts.isfinite):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(np.isfinite(np.asarray(values[0])).tolist(), elem=types.Bool),
        )

    @impl(stmts.isinf)
    def isinf(self, interp, frame: Frame, stmt: stmts.isinf):
        values = frame.get_values(stmt.args)
        return (ilist.IList(np.isinf(np.asarray(values[0])).tolist(), elem=types.Bool),)

    @impl(stmts.isnan)
    def isnan(self, interp, frame: Frame, stmt: stmts.isnan):
        values = frame.get_values(stmt.args)
        return (ilist.IList(np.isnan(np.asarray(values[0])).tolist(), elem=types.Bool),)

    @impl(stmts.lgamma)
    def lgamma(self, interp, frame: Frame, stmt: stmts.lgamma):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(
                special.loggamma(np.asarray(values[0])).tolist(), elem=types.Float
            ),
        )

    @impl(stmts.log10)
    def log10(self, interp, frame: Frame, stmt: stmts.log10):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(np.log10(np.asarray(values[0])).tolist(), elem=types.Float),
        )

    @impl(stmts.log1p)
    def log1p(self, interp, frame: Frame, stmt: stmts.log1p):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(np.log1p(np.asarray(values[0])).tolist(), elem=types.Float),
        )

    @impl(stmts.log2)
    def log2(self, interp, frame: Frame, stmt: stmts.log2):
        values = frame.get_values(stmt.args)
        return (ilist.IList(np.log2(np.asarray(values[0])).tolist(), elem=types.Float),)

    @impl(stmts.mult)
    def mult(self, interp, frame: Frame, stmt: stmts.mult):
        lhs = frame.get(stmt.lhs)
        rhs = frame.get(stmt.rhs)
        if isinstance(lhs, ilist.IList):
            lhs = np.asarray(lhs)
        if isinstance(rhs, ilist.IList):
            rhs = np.asarray(rhs)
        result = lhs * rhs
        return (ilist.IList(result.tolist(), elem=types.Float),)

    @impl(stmts.pow)
    def pow(self, interp, frame: Frame, stmt: stmts.pow):
        x = frame.get(stmt.x)
        y = frame.get(stmt.y)
        return (
            ilist.IList(
                np.power(np.asarray(x), y).tolist(),
                elem=types.Float,
            ),
        )

    @impl(stmts.radians)
    def radians(self, interp, frame: Frame, stmt: stmts.radians):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(np.radians(np.asarray(values[0])).tolist(), elem=types.Float),
        )

    @impl(stmts.remainder)
    def remainder(self, interp, frame: Frame, stmt: stmts.remainder):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(
                np.remainder(np.asarray(values[0]), np.asarray(values[1])).tolist(),
                elem=types.Float,
            ),
        )

    @impl(stmts.sin)
    def sin(self, interp, frame: Frame, stmt: stmts.sin):
        values = frame.get_values(stmt.args)
        return (ilist.IList(np.sin(np.asarray(values[0])).tolist(), elem=types.Float),)

    @impl(stmts.sinh)
    def sinh(self, interp, frame: Frame, stmt: stmts.sinh):
        values = frame.get_values(stmt.args)
        return (ilist.IList(np.sinh(np.asarray(values[0])).tolist(), elem=types.Float),)

    @impl(stmts.sqrt)
    def sqrt(self, interp, frame: Frame, stmt: stmts.sqrt):
        values = frame.get_values(stmt.args)
        return (ilist.IList(np.sqrt(np.asarray(values[0])).tolist(), elem=types.Float),)

    @impl(stmts.sub)
    def sub(self, interp, frame: Frame, stmt: stmts.sub):
        lhs = frame.get(stmt.lhs)
        rhs = frame.get(stmt.rhs)
        if isinstance(lhs, ilist.IList):
            lhs = np.asarray(lhs)
        if isinstance(rhs, ilist.IList):
            rhs = np.asarray(rhs)
        result = lhs - rhs
        return (ilist.IList(result.tolist(), elem=types.Float),)

    @impl(stmts.tan)
    def tan(self, interp, frame: Frame, stmt: stmts.tan):
        values = frame.get_values(stmt.args)
        return (ilist.IList(np.tan(np.asarray(values[0])).tolist(), elem=types.Float),)

    @impl(stmts.tanh)
    def tanh(self, interp, frame: Frame, stmt: stmts.tanh):
        values = frame.get_values(stmt.args)
        return (ilist.IList(np.tanh(np.asarray(values[0])).tolist(), elem=types.Float),)

    @impl(stmts.trunc)
    def trunc(self, interp, frame: Frame, stmt: stmts.trunc):
        values = frame.get_values(stmt.args)
        return (
            ilist.IList(np.trunc(np.asarray(values[0])).tolist(), elem=types.Float),
        )

    @impl(stmts.scale)
    def scale(self, interp, frame: Frame, stmt: stmts.scale):
        a = frame.get(stmt.value)
        x = frame.get(stmt.x)
        return (ilist.IList((np.asarray(x) * a).tolist(), elem=types.Float),)

    @impl(stmts.offset)
    def offset(self, interp, frame: Frame, stmt: stmts.offset):
        a = frame.get(stmt.value)
        x = frame.get(stmt.x)
        return (ilist.IList((np.asarray(x) + a).tolist(), elem=types.Float),)



from typing import Generic, TypeVar, TypeAlias, final
from dataclasses import dataclass

from kirin.ir import Block

ValueType = TypeVar("ValueType")


@final
@dataclass
class ReturnValue(Generic[ValueType]):
    """Return value from a statement evaluation.

    This class represents a return value from a statement evaluation. It is used
    to indicate that the statement evaluation should later pop the frame and
    return the value. Kirin does not allow multiple return values to follow Python
    semantics. If you want to return multiple values, you should return a tuple.
    """

    value: ValueType

    def __len__(self) -> int:
        return 0


@final
@dataclass
class YieldValue(Generic[ValueType]):
    """Yield value from a statement evaluation.

    This class represents values returned from a statement that terminates current
    region execution and returns the values to the caller. Unlike `ReturnValue`, this
    class won't pop the frame and return the value to the caller.
    """

    values: tuple[ValueType, ...]

    def __len__(self) -> int:
        return len(self.values)

    def __iter__(self):
        return iter(self.values)

    def __getitem__(self, index: int) -> ValueType:
        return self.values[index]


@final
@dataclass(init=False)
class Successor(Generic[ValueType]):
    """Successor block from a statement evaluation."""

    block: Block
    block_args: tuple[ValueType, ...]

    def __init__(self, block: Block, *block_args: ValueType):
        super().__init__()
        self.block = block
        self.block_args = block_args

    def __hash__(self) -> int:
        return hash(self.block)

    def __len__(self) -> int:
        return 0


SpecialValue: TypeAlias = (
    None | ReturnValue[ValueType] | YieldValue[ValueType] | Successor[ValueType]
)
"""Special value for statement evaluation."""
StatementResult: TypeAlias = tuple[ValueType, ...] | SpecialValue[ValueType]
"""Type alias for the result of a statement evaluation."""
RegionResult: TypeAlias = tuple[ValueType, ...] | None | ReturnValue[ValueType]
"""Type alias for the result of a region evaluation."""



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TypeVar
from contextlib import contextmanager
from dataclasses import field, dataclass

from kirin import ir
from kirin.interp import Frame, abc
from kirin.idtable import IdTable
from kirin.worklist import WorkList

TargetType = TypeVar("TargetType")


@dataclass
class EmitFrame(Frame[TargetType]):
    ssa: IdTable[ir.SSAValue] = field(
        default_factory=lambda: IdTable[ir.SSAValue](prefix="ssa_"),
        init=False,
    )
    block: IdTable[ir.Block] = field(
        default_factory=lambda: IdTable[ir.Block](prefix="block_"),
        init=False,
    )
    _indent: int = field(default=0, init=False)

    @contextmanager
    def indent(self):
        self._indent += 1
        try:
            yield
        finally:
            self._indent -= 1


CodeGenFrameType = TypeVar("CodeGenFrameType", bound=EmitFrame)


@dataclass
class EmitTable(IdTable[ir.Statement]):

    def add(self, value: ir.Statement) -> str:
        id = self.next_id
        if (trait := value.get_trait(ir.SymbolOpInterface)) is not None:
            value_name = trait.get_sym_name(value).unwrap()
            curr_ind = self.name_count.get(value_name, 0)
            suffix = f"_{curr_ind}" if curr_ind != 0 else ""
            self.name_count[value_name] = curr_ind + 1
            name = self.prefix + value_name + suffix
            self.table[value] = name
        else:
            name = f"{self.prefix}{self.prefix_if_none}{id}"
            self.next_id += 1
            self.table[value] = name
        return name

    def __getitem__(self, value: ir.Statement) -> str:
        if value in self.table:
            return self.table[value]
        raise KeyError(f"Symbol {value} not found in SymbolTable")

    def get(self, value: ir.Statement, default: str | None = None) -> str | None:
        if value in self.table:
            return self.table[value]
        return default


@dataclass
class EmitABC(abc.InterpreterABC[CodeGenFrameType, TargetType], ABC):
    callables: EmitTable = field(init=False)
    callable_to_emit: WorkList[ir.Statement] = field(init=False)

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        cls.callables = EmitTable(prefix="")
        cls.callable_to_emit = WorkList()
        for each in getattr(cls, "keys", ()):
            if not each.startswith("emit."):
                raise ValueError(f"Key {each} does not start with 'emit.'")

    def run(self, node: ir.Method | ir.Statement):
        self.reset()
        if isinstance(node, ir.Method):
            node = node.code

        with self.eval_context():
            self.callables.add(node)
            self.callable_to_emit.append(node)
            while self.callable_to_emit:
                callable = self.callable_to_emit.pop()
                if callable is None:
                    break
                self.eval(callable)
        return

    @abstractmethod
    def reset(self):
        """Reset any per-run state in the emitter."""
        pass



import pytest

from kirin.ir import Dialect
from kirin.types import PyClass


def test_py_type_register():

    class TestClass:
        pass

    class OtherTestClass:
        pass

    dialect = Dialect("test")

    TestType = dialect.register_py_type(
        TestClass, display_name="TestType", prefix="test"
    )
    assert TestType == PyClass(TestClass, prefix="test", display_name="TestType")

    assert dialect.python_types == {("test", "TestType"): TestType}

    with pytest.raises(ValueError):
        dialect.register_py_type(OtherTestClass, display_name="TestType", prefix="test")

    with pytest.raises(ValueError):
        dialect.register_py_type(TestClass, display_name="TestClass", prefix="test")

    with pytest.raises(ValueError):
        dialect.register_py_type(
            TestClass, display_name="TestType", prefix="other_prefix"
        )



import pytest

from kirin.prelude import basic_no_opt


@basic_no_opt
def no_return(x):
    return


def test_noreturn():
    assert no_return(1) is None

    with pytest.raises(ValueError):
        no_return()


def test_noreturn_with_body():
    @basic_no_opt
    def no_return_with_body(x):
        x + 1

    assert no_return_with_body(1) is None



"""Module dialect provides a simple module
that is roughly a list of function statements.

This dialect provides the dialect necessary for compiling a function into
lower-level IR with all its callee functions.
"""

from __future__ import annotations

from dataclasses import dataclass

from kirin import ir, types, interp
from kirin.decl import info, statement
from kirin.print import Printer
from kirin.analysis import TypeInference

dialect = ir.Dialect("module")


@dataclass(frozen=True)
class ModuleEntryPoint(ir.EntryPointInterface["Module"]):
    def get_entry_point_symbol(self, stmt: Module) -> str:
        return stmt.entry

    def get_entry_point(self, stmt: Module) -> ir.Statement:
        name = self.get_entry_point_symbol(stmt)
        for node in stmt.body.blocks[0].stmts:
            trait = node.get_trait(ir.SymbolOpInterface)
            if trait is None:
                continue

            node_name = trait.get_sym_name(node).unwrap()
            if node_name == name:
                return node
        raise ir.ValidationError(stmt, "entry point not found")


@statement(dialect=dialect)
class Module(ir.Statement):
    traits = frozenset(
        {
            ir.IsolatedFromAbove(),
            ir.SymbolTable(),
            ir.SymbolOpInterface(),
            ModuleEntryPoint(),
        }
    )
    sym_name: str = info.attribute()
    entry: str = info.attribute()
    body: ir.Region = info.region(multi=False)

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(" ")
        with printer.rich(style="symbol"):
            printer.plain_print("@", self.sym_name)
        printer.plain_print(" {")
        with printer.indent():
            for idx, stmt in enumerate(self.body.blocks[0].stmts):
                printer.print_newline()
                stmt.print(printer)
        printer.print_newline()
        printer.plain_print("}")
        with printer.rich(style="comment"):
            printer.plain_print(" // entry: ", self.entry)


@statement(dialect=dialect)
class Invoke(ir.Statement):
    """A special statement that represents
    a function calling functions by symbol name.

    Note:
        This statement is here for completeness, for interpretation,
        it is recommended to rewrite this statement into a `func.Invoke`
        after looking up the symbol table.
    """

    traits = frozenset({ir.MaybePure()})
    callee: str = info.attribute()
    inputs: tuple[ir.SSAValue, ...] = info.argument()
    kwargs: tuple[ir.SSAValue, ...] = info.argument()
    keys: tuple[str, ...] = info.attribute(default=())
    result: ir.ResultValue = info.result()
    purity: bool = info.attribute(default=False)

    def print_impl(self, printer: Printer) -> None:
        with printer.rich(style="red"):
            printer.print_name(self)
        printer.plain_print(" ")

        with printer.rich(style="symbol"):
            printer.plain_print("@", self.callee)
        kwargs = dict(zip(self.keys, self.kwargs))
        printer.plain_print("(")
        printer.print_seq(self.inputs)
        if kwargs and self.inputs:
            printer.plain_print(", ")
        printer.print_mapping(kwargs, delim=", ")
        printer.plain_print(")")

        with printer.rich(style="comment"):
            printer.plain_print(" : ")
            printer.print_seq(
                [result.type for result in self._results],
                delim=", ",
            )
            printer.plain_print(f" maybe_pure={self.purity}")

    def check(self) -> None:
        assert len(self.keys) == len(
            self.kwargs
        ), "keys and kwargs must have the same length"


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(Invoke)
    def interp_Invoke(
        self, interp_: interp.Interpreter, frame: interp.Frame, stmt: Invoke
    ):
        callee = interp_.symbol_table.get(stmt.callee)
        if callee is None:
            raise interp.InterpreterError(f"symbol {stmt.callee} not found")

        _, ret = interp_.call(
            callee, ir.Method(interp_.dialects, callee), *frame.get_values(stmt.inputs)
        )
        return (ret,)


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(Invoke)
    def typeinfer_Invoke(
        self, interp_: TypeInference, frame: interp.Frame, stmt: Invoke
    ):
        callee = interp_.symbol_table.get(stmt.callee)
        if callee is None:
            return (types.Bottom,)

        callee = interp_.symbol_table.get(stmt.callee)
        if callee is None:
            raise interp.InterpreterError(f"symbol {stmt.callee} not found")

        _, ret = interp_.call(
            callee,
            interp_.method_self(ir.Method(interp_.dialects, callee)),
            *frame.get_values(stmt.inputs),
        )
        return (ret,)



from kirin import ir, types, lowering
from kirin.decl import info, statement

from ._dialect import dialect


@statement(dialect=dialect)
class Random(ir.Statement):
    """random statement, wrapping the random.random function
    returns a random floating number between 0 and 1
    """

    traits = frozenset({lowering.FromPythonCall()})
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class RandInt(ir.Statement):
    """randint statement, wrapping the random.randint function
    returns a random integer between the given range
    """

    traits = frozenset({lowering.FromPythonCall()})
    start: ir.SSAValue = info.argument(types.Int)
    stop: ir.SSAValue = info.argument(types.Int)
    result: ir.ResultValue = info.result(types.Int)


@statement(dialect=dialect)
class Uniform(ir.Statement):
    """uniform statement, wrapping the random.uniform function
    returns a random floating number between the given range
    """

    traits = frozenset({lowering.FromPythonCall()})
    start: ir.SSAValue = info.argument(types.Float)
    stop: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class Seed(ir.Statement):
    """seed statement, wrapping the random.seed function
    sets the seed for the random number generator
    """

    traits = frozenset({lowering.FromPythonCall()})
    value: ir.SSAValue = info.argument(types.Int)



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeVar,
    Callable,
    Optional,
    cast,
    overload,
)
from dataclasses import field, dataclass

from kirin.ir import Block, Region, SSAValue, Statement

from .stream import Stmt, StmtStream
from .exception import BuildError

if TYPE_CHECKING:
    from .state import State

CallbackFn = Callable[["Frame", SSAValue], SSAValue]
StmtType = TypeVar("StmtType", bound=Statement)


@dataclass
class Frame(Generic[Stmt]):
    state: State
    """lowering state"""
    parent: Optional[Frame] = field(default=None, init=False)
    """parent frame, if any"""
    stream: StmtStream[Stmt]
    """stream of statements"""

    curr_region: Region
    """the region this frame is generating"""
    entr_block: Block
    """entry block of the frame region"""
    curr_block: Block
    """current block being lowered"""
    next_block: Block
    """next block to be lowered, but not yet inserted in the region"""

    defs: dict[str, SSAValue] = field(default_factory=dict)
    """values defined in the current frame"""
    globals: dict[str, Any] = field(default_factory=dict)
    """global values known to the current frame"""
    captures: dict[str, SSAValue] = field(default_factory=dict)
    """values accessed from the parent frame"""
    capture_callback: Optional[CallbackFn] = None
    """callback function that creates a local SSAValue value when an captured value was used."""

    def __repr__(self):
        return f"Frame({len(self.defs)} defs, {len(self.globals)} globals)"

    @overload
    def push(self, node: StmtType) -> StmtType: ...

    @overload
    def push(self, node: Block) -> Block: ...

    def push(self, node: StmtType | Block) -> StmtType | Block:
        if node.IS_BLOCK:
            return self._push_block(cast(Block, node))
        elif node.IS_STATEMENT:
            return self._push_stmt(cast(StmtType, node))
        else:
            raise BuildError(f"Unsupported type {type(node)} in push()")

    def _push_stmt(self, stmt: StmtType) -> StmtType:
        if not stmt.dialect:
            raise BuildError(f"unexpected builtin statement {stmt.name}")
        elif stmt.dialect not in self.state.parent.dialects:
            raise BuildError(
                f"Unsupported dialect `{stmt.dialect.name}` from statement {stmt.name}"
            )
        if stmt.source is None:
            stmt.source = self.state.source
        self.curr_block.stmts.append(stmt)
        return stmt

    def _push_block(self, block: Block):
        """Append a block to the current region.

        Args:
            block(Block): block to append, default `None` to create a new block.
        """
        self.curr_region.blocks.append(block)
        self.curr_block = block
        if block.source is None:
            block.source = self.state.source
        return block

    def jump_next_block(self):
        """Jump to the next block and return it.
        This appends the current `Frame.next_block` to the current region
        and creates a new Block for `next_block`.

        Returns:
            Block: the next block
        """
        block = self.push(self.next_block)
        self.next_block = Block()
        return block

    def get(self, name: str) -> SSAValue | None:
        value = self.get_local(name)
        if value is not None:
            return value

        # NOTE: look up local first, then globals
        if name in self.globals:
            return self.state.get_literal(self.globals[name])
        return None

    def get_local(self, name: str) -> SSAValue | None:
        if name in self.defs:
            return self.defs[name]

        if self.parent is None:
            return None  # no parent frame, return None

        value = self.parent.get_local(name)
        if value is not None:
            self.captures[name] = value
            if self.capture_callback:
                # whatever generates a local value gets defined
                ret = self.capture_callback(self, value)
                self.defs[name] = ret
                return ret
            return value
        return None

    def __getitem__(self, name: str) -> SSAValue:
        """Get a variable from current scope.

        Args:
            name(str): variable name

        Returns:
            SSAValue: the value of the variable

        Raises:
            lowering.BuildError: if the variable is not found in the scope,
                or if the variable has multiple possible values.
        """
        value = self.defs.get(name)
        if isinstance(value, SSAValue):
            return value
        else:
            raise BuildError(f"Variable {name} not found in scope")

    def exhaust(self):
        """Exhaust the current stream and return the remaining statements."""
        stream = self.stream
        while stream:
            stmt = stream.pop()
            self.state.parent.visit(self.state, stmt)



from typing import cast
from dataclasses import field, dataclass

from kirin import ir
from kirin.rewrite.abc import RewriteRule, RewriteResult

from ..cf import Branch, ConditionalBranch
from ..func import ConstantNone
from .stmts import For, Yield, IfElse
from ..py.cmp import Is
from ..py.iterable import Iter, Next


class ScfRule(RewriteRule):

    def get_entr_and_exit_blks(self, node: For | IfElse):
        """Get the enter and exit blocks for the given SCF node.

        The exit block is a new block that will be created to hold the
        statements that follow the SCF node in the current block and the
        enter block is a new block that will be created to hold the
        any logic required to enter the SCF node.

        """
        # split the current block into two parts
        exit_block = ir.Block()
        stmt = node.next_stmt
        while stmt is not None:
            next_stmt = stmt.next_stmt
            stmt.detach()
            exit_block.stmts.append(stmt)
            stmt = next_stmt

        for result in node.results:
            result.replace_by(exit_block.args.append_from(result.type, result.name))

        curr_block = node.parent_block
        assert (
            curr_block is not None and curr_block.IS_BLOCK
        ), "Node must be inside a block"
        curr_block = cast(ir.Block, curr_block)

        curr_block.stmts.append(
            Branch(arguments=(), successor=(entr_block := ir.Block()))
        )

        return exit_block, entr_block

    def get_curr_blk_info(self, node: For | IfElse) -> tuple[ir.Region, int]:
        """Get the current region and the block index of the node in the region."""
        curr_block = node.parent_block
        region = node.parent_region

        assert region is not None and region.IS_REGION, "Node must be inside a region"
        region = cast(ir.Region, region)
        assert (
            curr_block is not None and curr_block.IS_BLOCK
        ), "Node must be inside a block"
        curr_block = cast(ir.Block, curr_block)

        block_idx = region._block_idx[curr_block]
        return region, block_idx


class ForRule(ScfRule):
    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if (
            not isinstance(node, For)
            # must be inside a callable statement
            or not isinstance(parent_stmt := node.parent_stmt, ir.Statement)
            or not parent_stmt.has_trait(ir.CallableStmtInterface)
        ):
            return RewriteResult()

        region, block_idx = self.get_curr_blk_info(node)
        exit_block, entr_block = self.get_entr_and_exit_blks(node)

        (body_block := node.body.blocks[0]).detach()

        # Get iterator from iterable object
        entr_block.stmts.append(iterable_stmt := Iter(node.iterable))
        entr_block.stmts.append(next_stmt := Next(iterable_stmt.expect_one_result()))
        entr_block.stmts.append(const_none := ConstantNone())
        entr_block.stmts.append(
            loop_cmp := Is(next_stmt.expect_one_result(), const_none.result)
        )
        entr_block.stmts.append(
            ConditionalBranch(
                cond=loop_cmp.result,
                then_arguments=tuple(node.initializers),
                then_successor=exit_block,
                else_arguments=(next_stmt.expect_one_result(),)
                + tuple(node.initializers),
                else_successor=body_block,
            )
        )

        if isinstance(last_stmt := body_block.last_stmt, Yield):
            (next_stmt := Next(iterable_stmt.expect_one_result())).insert_before(
                last_stmt
            )
            (const_none := ConstantNone()).insert_before(last_stmt)
            (
                loop_cmp := Is(next_stmt.expect_one_result(), const_none.result)
            ).insert_before(last_stmt)
            last_stmt.replace_by(
                ConditionalBranch(
                    cond=loop_cmp.result,
                    else_arguments=(next_stmt.expect_one_result(),)
                    + tuple(last_stmt.args),
                    else_successor=body_block,
                    then_arguments=tuple(last_stmt.args),
                    then_successor=exit_block,
                )
            )

        # insert the body block and add branch to the current block
        region.blocks.insert(block_idx + 1, exit_block)
        region.blocks.insert(block_idx + 1, body_block)
        region.blocks.insert(block_idx + 1, entr_block)

        node.delete()

        return RewriteResult(has_done_something=True)


class IfElseRule(ScfRule):
    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if (
            not isinstance(node, IfElse)
            or not isinstance(parent_stmt := node.parent_stmt, ir.Statement)
            or not parent_stmt.has_trait(ir.CallableStmtInterface)
        ):
            return RewriteResult()

        region, block_idx = self.get_curr_blk_info(node)
        exit_block, entr_block = self.get_entr_and_exit_blks(node)

        (then_block := node.then_body.blocks[0]).detach()
        (else_block := node.else_body.blocks[0]).detach()
        entr_block.stmts.append(
            ConditionalBranch(
                node.cond,
                then_arguments=tuple(node.args),
                then_successor=then_block,
                else_arguments=tuple(node.args),
                else_successor=else_block,
            )
        )

        if isinstance(last_stmt := then_block.last_stmt, Yield):
            last_stmt.replace_by(
                Branch(
                    arguments=tuple(last_stmt.args),
                    successor=exit_block,
                )
            )

        if isinstance(last_stmt := else_block.last_stmt, Yield):
            last_stmt.replace_by(
                Branch(
                    arguments=tuple(last_stmt.args),
                    successor=exit_block,
                )
            )

        # insert the new blocks
        region.blocks.insert(block_idx + 1, exit_block)
        region.blocks.insert(block_idx + 1, else_block)
        region.blocks.insert(block_idx + 1, then_block)
        region.blocks.insert(block_idx + 1, entr_block)

        node.delete()
        return RewriteResult(has_done_something=True)


@dataclass
class ScfToCfRule(RewriteRule):

    for_rule: ForRule = field(default_factory=ForRule, init=False)
    if_else_rule: IfElseRule = field(default_factory=IfElseRule, init=False)

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if isinstance(node, For):
            return self.for_rule.rewrite_Statement(node)
        elif isinstance(node, IfElse):
            return self.if_else_rule.rewrite_Statement(node)
        else:
            return RewriteResult()



from abc import ABC
from typing import cast
from dataclasses import field, dataclass

from kirin.ir import Pure, Block, IRNode, Region, MaybePure, Statement


@dataclass
class RewriteResult:
    terminated: bool = field(default=False, kw_only=True)
    has_done_something: bool = field(default=False, kw_only=True)
    exceeded_max_iter: bool = field(default=False, kw_only=True)

    def join(self, other: "RewriteResult") -> "RewriteResult":
        return RewriteResult(
            terminated=self.terminated or other.terminated,
            has_done_something=self.has_done_something or other.has_done_something,
            exceeded_max_iter=self.exceeded_max_iter or other.exceeded_max_iter,
        )


@dataclass(repr=False)
class RewriteRule(ABC):
    """A rewrite rule that matches and rewrites IR nodes.

    The rewrite rule is applied to an IR node by calling the instance with the node as an argument.
    The rewrite rule should mutate the node instead of returning a new node. A `RewriteResult` should
    be returned to indicate whether the rewrite rule has done something, whether the rewrite rule
    should terminate, and whether the rewrite rule has exceeded the maximum number of iterations.
    """

    def rewrite(self, node: IRNode) -> RewriteResult:
        if node.IS_REGION:
            return self.rewrite_Region(cast(Region, node))
        elif node.IS_BLOCK:
            return self.rewrite_Block(cast(Block, node))
        elif node.IS_STATEMENT:
            return self.rewrite_Statement(cast(Statement, node))
        else:
            return RewriteResult()

    def rewrite_Region(self, node: Region) -> RewriteResult:
        return RewriteResult()

    def rewrite_Block(self, node: Block) -> RewriteResult:
        return RewriteResult()

    def rewrite_Statement(self, node: Statement) -> RewriteResult:
        return RewriteResult()

    def is_pure(self, node: Statement):
        if node.has_trait(Pure):
            return True

        if (trait := node.get_trait(MaybePure)) and trait.is_pure(node):
            return True
        return False



from kirin import ir, types
from kirin.analysis import const
from kirin.rewrite.abc import RewriteRule, RewriteResult
from kirin.dialects.py.constant import Constant

from ..stmts import IListType
from ..runtime import IList
from .._dialect import dialect


@dialect.post_inference
class ConstList2IList(RewriteRule):
    """Rewrite type annotation for SSAValue with constant `IList`
    in `Hinted` type. This should be run after constant folding and
    `WrapConst` rule.
    """

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if isinstance(node, Constant):
            return self.rewrite_Constant(node)

        has_done_something = False
        for result in node.results:
            if not isinstance(hint := result.hints.get("const"), const.Value):
                continue

            typ = result.type
            data = hint.data
            if isinstance(typ, types.PyClass) and typ.is_subseteq(types.PyClass(IList)):
                has_done_something = has_done_something or self._rewrite_IList_type(
                    result, data
                )
            elif isinstance(typ, types.Generic) and typ.body.is_subseteq(
                types.PyClass(IList)
            ):
                has_done_something = has_done_something or self._rewrite_IList_type(
                    result, data
                )
        return RewriteResult(has_done_something=has_done_something)

    def rewrite_Constant(self, node: Constant) -> RewriteResult:
        if not isinstance(node.value, ir.PyAttr):
            return RewriteResult()

        if isinstance(data := node.value.data, list):
            stmt = Constant(value=IList(data=data))
            node.replace_by(stmt)
            self._rewrite_IList_type(stmt.result, data)
            return RewriteResult(has_done_something=True)
        elif isinstance(data, range):
            new_constant = IList(data=data, elem=types.Int)
            stmt = Constant(value=new_constant)
            # specializing the type computation since we know that a
            # range will always be integer typed.
            stmt.result.hints["const"] = const.Value(new_constant)
            stmt.result.type = IListType[types.Int, types.Literal(len(data))]
            node.replace_by(stmt)
            return RewriteResult(has_done_something=True)

        return RewriteResult()

    def _rewrite_IList_type(self, result: ir.SSAValue, data):
        if not isinstance(data, IList):
            return False

        if not data.data:
            return False

        elem_type = types.PyClass(type(data[0]))
        for elem in data.data[1:]:
            elem_type = elem_type.join(types.PyClass(type(elem)))

        new_type = IListType[elem_type, types.Literal(len(data.data))]
        new_hint = const.Value(data)

        # Check if type and hint are already correct
        if result.type == new_type and result.hints.get("const") == new_hint:
            return False

        result.type = new_type
        result.hints["const"] = new_hint
        return True



from dataclasses import dataclass

from kirin.prelude import basic


def test_getattr():
    @dataclass
    class MyFoo:
        x: float
        y: float
        z: float

    foo = MyFoo(1.0, 2.0, 3.0)

    @basic
    def main():
        return foo.x + 1.0

    main.print()
    out = main()

    assert out == 2.0



from pytest import mark

from kirin import types
from kirin.prelude import structural_no_opt
from kirin.analysis import TypeInference

type_infer = TypeInference(structural_no_opt)


@mark.xfail(reason="for with early return not supported in scf lowering")
def test_inside_return_loop():
    @structural_no_opt
    def simple_loop(x: float):
        for i in range(0, 3):
            return i
        return x

    frame, ret = type_infer.run(simple_loop)
    assert ret.is_subseteq(types.Int | types.Float)


@mark.xfail(reason="if with early return not supported in scf lowering")
def test_simple_ifelse():
    @structural_no_opt
    def simple_ifelse(x: int):
        cond = x > 0
        if cond:
            return cond
        else:
            return 0

    frame, ret = type_infer.run(simple_ifelse)
    assert ret.is_subseteq(types.Bool | types.Int | types.NoneType)



from __future__ import annotations

import ast
from abc import abstractmethod
from typing import Any, Generic, TypeVar, get_origin
from dataclasses import dataclass

from kirin import ir
from kirin.decl import fields
from kirin.lowering.abc import Result
from kirin.lowering.state import State
from kirin.lowering.exception import BuildError

ASTNode = TypeVar("ASTNode", bound=ast.AST)
StmtType = TypeVar("StmtType", bound="ir.Statement")


@dataclass(frozen=True)
class PythonLoweringTrait(ir.Trait[ir.Statement], Generic[StmtType, ASTNode]):
    """A trait that indicates that a statement can be lowered from Python AST."""

    @abstractmethod
    def lower(
        self, stmt: type[StmtType], state: State[ast.AST], node: ASTNode
    ) -> Result: ...

    @classmethod
    def lower_Call_inputs(
        cls, stmt: type[StmtType], state: State[ast.AST], node: ast.Call
    ) -> tuple[dict[str, ir.SSAValue | tuple[ir.SSAValue, ...]], dict[str, Any]]:
        """Lower the inputs of a Python call to corresponding SSA values or
        compile-time values (attributes).

        Args:
            stmt: The statement class to lower to.
            state: The lowering state.
            node: The Python call node to lower.

        Returns:
            A tuple containing two dictionaries:
            - The first dictionary contains the standard arguments and their values.
            - The second dictionary contains the keyword arguments and their values.

        Raises:
            lowering.BuildError: If the Python call cannot be lowered to the statement.
        """
        fs = fields(stmt)
        stmt_std_arg_names = fs.std_args.keys()
        stmt_kw_args_name = fs.kw_args.keys()
        stmt_attr_prop_names = fs.attr_or_props
        stmt_required_names = fs.required_names
        stmt_group_arg_names = fs.group_arg_names
        args, kwargs = {}, {}
        for name, value in zip(stmt_std_arg_names, node.args):
            cls.__parse_arg(state, stmt_group_arg_names, args, name, value)
        for kw in node.keywords:
            if not isinstance(kw.arg, str):
                raise BuildError("Expected string for keyword argument name")

            arg: str = kw.arg
            if arg in node.args:
                raise BuildError(
                    f"Keyword argument {arg} is already present in positional arguments"
                )
            elif arg in stmt_std_arg_names or arg in stmt_kw_args_name:
                cls.__parse_arg(state, stmt_group_arg_names, kwargs, kw.arg, kw.value)
            elif arg in stmt_attr_prop_names:
                if (
                    isinstance(kw.value, ast.Name)
                    and state.current_frame.get_local(kw.value.id) is not None
                ):
                    raise BuildError(
                        f"Expected global/constant value for attribute or property {arg}"
                    )
                global_value = state.get_global(kw.value)
                if (decl := fs.attributes.get(arg)) is not None:
                    if decl.annotation is Any:
                        kwargs[arg] = global_value.data
                    else:
                        kwargs[arg] = global_value.expect(
                            get_origin(decl.annotation) or decl.annotation
                        )
                else:
                    raise BuildError(f"Unexpected attribute or property {arg}")
            else:
                raise BuildError(f"Unexpected keyword argument {arg}")

        for name in stmt_required_names:
            if name not in args and name not in kwargs:
                raise BuildError(f"Missing required argument {name}")

        return args, kwargs

    @classmethod
    def __parse_arg(
        cls,
        state: State[ast.AST],
        group_names: set[str],
        target: dict,
        name: str,
        value: ast.AST,
    ):
        if name in group_names:
            if not isinstance(value, ast.Tuple):
                raise BuildError(f"Expected tuple for group argument {name}")
            target[name] = tuple(state.lower(elem).expect_one() for elem in value.elts)
        else:
            target[name] = state.lower(value).expect_one()


StatementType = TypeVar("StatementType", bound="ir.Statement")


@dataclass(frozen=True)
class FromPythonCall(PythonLoweringTrait[StatementType, ast.Call]):
    """Trait for customizing lowering of Python calls to a statement.

    Declared in a statement definition to indicate that the statement can be
    constructed from a Python call (i.e., a function call `ast.Call` in the
    Python AST).

    Subclassing this trait allows for customizing the lowering of Python calls
    to the statement. The `lower` method should be implemented to parse the
    arguments from the Python call and construct the statement instance.
    """

    def lower(
        self, stmt: type[StatementType], state: State[ast.AST], node: ast.Call
    ) -> Result:
        args, kwargs = self.lower_Call_inputs(stmt, state, node)
        return state.current_frame.push(stmt(*args.values(), **kwargs))

    def verify(self, node: ir.Statement):
        assert len(node.regions) == 0, "FromPythonCall statements cannot have regions"
        assert (
            len(node.successors) == 0
        ), "FromPythonCall statements cannot have successors"


@dataclass(frozen=True)
class FromPythonRangeLike(FromPythonCall[StatementType]):
    """Provides a default lowering implementation for built-in `range`-like function
    to a statement that takes three arguments: start, stop, and step.
    """

    def lower(
        self, stmt: type[StatementType], state: State[ast.AST], node: ast.Call
    ) -> Result:
        nargs = len(node.args)
        if nargs == 1:
            start = state.get_literal(0)
            stop = state.lower(node.args[0]).expect_one()
            step = state.get_literal(1)
        elif nargs == 2:
            start = state.lower(node.args[0]).expect_one()
            stop = state.lower(node.args[1]).expect_one()
            step = state.get_literal(1)
        elif nargs == 3:
            start = state.lower(node.args[0]).expect_one()
            stop = state.lower(node.args[1]).expect_one()
            step = state.lower(node.args[2]).expect_one()
        else:
            raise BuildError("range() takes 1-3 arguments")

        return state.current_frame.push(stmt(start, stop, step))  # type: ignore


@dataclass(frozen=True)
class FromPythonWith(PythonLoweringTrait[StatementType, ast.With]):
    """Trait for customizing lowering of Python with statements to a statement.

    Subclassing this trait allows for customizing the lowering of Python with
    statements to the statement. The `lower` method should be implemented to parse
    the arguments from the Python with statement and construct the statement instance.
    """

    pass


@dataclass(frozen=True)
class FromPythonWithSingleItem(FromPythonWith[StatementType]):
    """Trait for customizing lowering of the following Python with syntax to a statement:

    ```python
    with <stmt>[ as <name>]:
        <body>
    ```

    where `<stmt>` is the statement being lowered, `<name>` is an optional name for the result
    of the statement, and `<body>` is the body of the with statement. The optional `as <name>`
    is not valid when the statement has no results.

    This syntax is slightly different from the standard Python `with` statement in that
    `<name>` refers to the result of the statement, not the context manager. Thus typically
    one sould access `<name>` in `<body>` to use the result of the statement.

    In some cases, however, `<name>` may be used as a reference of a special value `self` that
    is passed to the `<body>` of the statement. This is useful for statements that have a similar
    behavior to a closure.
    """

    def lower(
        self, stmt: type[StatementType], state: State[ast.AST], node: ast.With
    ) -> Result:
        from kirin.dialects import cf

        fs = fields(stmt)
        if len(fs.regions) != 1:
            raise BuildError("Expected exactly one region in statement declaration")

        if len(node.items) != 1:
            raise BuildError("Expected exactly one item in statement")

        item, body = node.items[0], node.body
        if not isinstance(item.context_expr, ast.Call):
            raise BuildError(
                f"Expected context expression to be a call for with {stmt.name}"
            )

        region_name, region_info = next(iter(fs.regions.items()))
        if region_info.multi:  # branch to exit block if not terminated
            with state.frame(body) as body_frame:
                body_frame.exhaust()
                for block in body_frame.curr_region.blocks:
                    if block.last_stmt is None or not block.last_stmt.has_trait(
                        ir.IsTerminator
                    ):
                        block.stmts.append(
                            cf.Branch(arguments=(), successor=body_frame.next_block)
                        )
                for block in body_frame.curr_region.blocks:
                    if block.last_stmt is None or not block.last_stmt.has_trait(
                        ir.IsTerminator
                    ):
                        block.stmts.append(
                            cf.Branch(arguments=(), successor=body_frame.next_block)
                        )
        else:
            with state.frame(body, finalize_next=False) as body_frame:
                body_frame.exhaust()
                if len(body_frame.curr_region.blocks) != 1:
                    raise BuildError(
                        f"Expected exactly one block in region {region_name}"
                    )

                if len(body_frame.curr_region.blocks) != 1:
                    raise BuildError(
                        f"Expected exactly one block in region {region_name}"
                    )

        args, kwargs = self.lower_Call_inputs(stmt, state, item.context_expr)
        kwargs[region_name] = body_frame.curr_region
        results = state.current_frame.push(stmt(*args.values(), **kwargs)).results
        if len(results) == 0:
            return
        elif len(results) > 1:
            raise BuildError(
                f"Expected exactly one result or no result from statement {stmt.name}"
            )

        result = results[0]
        if item.optional_vars is not None and isinstance(item.optional_vars, ast.Name):
            result.name = item.optional_vars.id
            state.current_frame.defs[result.name] = result
        return

    def verify(self, node: ir.Statement):
        assert (
            len(node.regions) == 1
        ), "FromPythonWithSingleItem statements must have one region"
        assert (
            len(node.successors) == 0
        ), "FromPythonWithSingleItem statements cannot have successors"
        assert (
            len(node.results) <= 1
        ), "FromPythonWithSingleItem statements can have at most one result"



from pytest import mark

from kirin import types
from kirin.prelude import structural_no_opt
from kirin.analysis import TypeInference


@mark.xfail(reason="if with early return not supported in scf lowering")
def test_self_ref_closure():

    @structural_no_opt
    def should_work(n_qubits: int):
        def self_ref_source(i_layer):
            stride = n_qubits // (2**i_layer)
            if stride == 0:
                return

            self_ref_source(i_layer + 1)

        return self_ref_source

    infer = TypeInference(structural_no_opt)
    frame, ret = infer.run(should_work, types.Int)
    should_work.print(analysis=frame.entries)
    assert ret.is_structurally_equal(
        types.MethodType[types.Tuple[types.Any], types.NoneType]
    )



"""Lattice for constant analysis."""

from typing import Any, final
from dataclasses import dataclass

from kirin import ir
from kirin.lattice import (
    BoundedLattice,
    IsSubsetEqMixin,
    SimpleJoinMixin,
    SimpleMeetMixin,
)
from kirin.ir.attrs.abc import LatticeAttributeMeta, SingletonLatticeAttributeMeta
from kirin.print.printer import Printer

from ._visitor import _ElemVisitor


@dataclass
class Result(
    ir.Attribute,
    IsSubsetEqMixin["Result"],
    SimpleJoinMixin["Result"],
    SimpleMeetMixin["Result"],
    BoundedLattice["Result"],
    _ElemVisitor,
    metaclass=LatticeAttributeMeta,
):
    """Base class for constant analysis results."""

    @classmethod
    def top(cls) -> "Result":
        return Unknown()

    @classmethod
    def bottom(cls) -> "Result":
        return Bottom()

    def print_impl(self, printer: Printer) -> None:
        printer.plain_print(repr(self))


@final
@dataclass
class Unknown(Result, metaclass=SingletonLatticeAttributeMeta):
    """Unknown constant value. This is the top element of the lattice."""

    def is_subseteq(self, other: Result) -> bool:
        return isinstance(other, Unknown)

    def __hash__(self) -> int:
        return id(self)

    def is_structurally_equal(
        self, other: ir.Attribute, context: dict | None = None
    ) -> bool:
        return isinstance(other, Unknown)


@final
@dataclass
class Bottom(Result, metaclass=SingletonLatticeAttributeMeta):
    """Bottom element of the lattice."""

    def is_subseteq(self, other: Result) -> bool:
        return True

    def __hash__(self) -> int:
        return id(self)

    def is_structurally_equal(
        self, other: ir.Attribute, context: dict | None = None
    ) -> bool:
        return isinstance(other, Bottom)


@final
@dataclass
class Value(Result):
    """Constant value. Wraps any hashable Python value."""

    data: Any

    def is_subseteq_Value(self, other: "Value") -> bool:
        return self.data == other.data

    def __hash__(self) -> int:
        # NOTE: we use id here because the data
        # may not be hashable. This is fine because
        # the data is guaranteed to be unique.
        return id(self)

    def is_structurally_equal(
        self, other: ir.Attribute, context: dict | None = None
    ) -> bool:
        if not isinstance(other, Value):
            return False
        return self.data == other.data


@dataclass
class PartialConst(Result):
    """Base class for partial constant values."""

    pass


@final
class PartialTupleMeta(LatticeAttributeMeta):
    """Metaclass for PartialTuple.

    This metaclass canonicalizes PartialTuple instances with all Value elements
    into a single Value instance.
    """

    def __call__(cls, data: tuple[Result, ...]):
        if all(isinstance(x, Value) for x in data):
            return Value(tuple(x.data for x in data))  # type: ignore
        return super().__call__(data)


@final
@dataclass
class PartialTuple(PartialConst, metaclass=PartialTupleMeta):
    """Partial tuple constant value."""

    data: tuple[Result, ...]

    def join(self, other: Result) -> Result:
        if other.is_subseteq(self):
            return self
        elif self.is_subseteq(other):
            return other
        elif isinstance(other, PartialTuple):
            return PartialTuple(tuple(x.join(y) for x, y in zip(self.data, other.data)))
        elif isinstance(other, Value) and isinstance(other.data, tuple):
            return PartialTuple(
                tuple(x.join(Value(y)) for x, y in zip(self.data, other.data))
            )
        return Unknown()

    def meet(self, other: Result) -> Result:
        if self.is_subseteq(other):
            return self
        elif other.is_subseteq(self):
            return other
        elif isinstance(other, PartialTuple):
            return PartialTuple(tuple(x.meet(y) for x, y in zip(self.data, other.data)))
        elif isinstance(other, Value) and isinstance(other.data, tuple):
            return PartialTuple(
                tuple(x.meet(Value(y)) for x, y in zip(self.data, other.data))
            )
        return self.bottom()

    def is_subseteq_PartialTuple(self, other: "PartialTuple") -> bool:
        return all(x.is_subseteq(y) for x, y in zip(self.data, other.data))

    def is_subseteq_Value(self, other: Value) -> bool:
        if isinstance(other.data, tuple):
            return all(x.is_subseteq(Value(y)) for x, y in zip(self.data, other.data))
        return False

    def __hash__(self) -> int:
        return hash(self.data)

    def is_structurally_equal(
        self, other: ir.Attribute, context: dict | None = None
    ) -> bool:
        if isinstance(other, PartialTuple):
            return all(
                x.is_structurally_equal(y, context=context)
                for x, y in zip(self.data, other.data)
            )
        elif isinstance(other, Value) and isinstance(other.data, tuple):
            return all(
                x.is_structurally_equal(y, context=context)
                for x, y in zip(self.data, other.data)
            )
        return False


@final
@dataclass
class PartialLambda(PartialConst):
    """Partial lambda constant value.

    This represents a closure with captured variables.
    """

    code: ir.Statement
    captured: tuple[Result, ...]
    argnames: list[str]

    def __init__(
        self,
        code: ir.Statement,
        captured: tuple[Result, ...],
        argnames: list[str] | None = None,
    ):
        trait = code.get_present_trait(ir.CallableStmtInterface)
        region = trait.get_callable_region(code)
        self.argnames = argnames or [
            arg.name or f"arg_{idx}" for idx, arg in enumerate(region.blocks[0].args)
        ]
        self.code = code
        self.captured = captured

    def __hash__(self) -> int:
        return hash((self.argnames, self.code, self.captured))

    def is_subseteq_PartialLambda(self, other: "PartialLambda") -> bool:
        if self.code is not other.code:
            return False
        if len(self.captured) != len(other.captured):
            return False

        return all(x.is_subseteq(y) for x, y in zip(self.captured, other.captured))

    def join(self, other: Result) -> Result:
        if other is other.bottom():
            return self

        if not isinstance(other, PartialLambda):
            return Unknown().join(other)  # widen self

        if self.code is not other.code:
            return Unknown()  # lambda stmt is pure

        if len(self.captured) != len(other.captured):
            return self.bottom()  # err

        return PartialLambda(
            self.code,
            tuple(x.join(y) for x, y in zip(self.captured, other.captured)),
            self.argnames,
        )

    def meet(self, other: Result) -> Result:
        if not isinstance(other, PartialLambda):
            return Unknown().meet(other)

        if self.code is not other.code:
            return self.bottom()

        if len(self.captured) != len(other.captured):
            return Unknown()

        return PartialLambda(
            self.code,
            tuple(x.meet(y) for x, y in zip(self.captured, other.captured)),
            self.argnames,
        )

    def is_structurally_equal(
        self, other: ir.Attribute, context: dict | None = None
    ) -> bool:
        return (
            isinstance(other, PartialLambda)
            and self.code.is_structurally_equal(other.code, context=context)
            and self.argnames == other.argnames
            and len(self.captured) == len(other.captured)
            and all(
                x.is_structurally_equal(y, context=context)
                for x, y in zip(self.captured, other.captured)
            )
        )



from typing import TYPE_CHECKING, Any, Type, TypeVar
from dataclasses import dataclass

from typing_extensions import Protocol, runtime_checkable

from kirin.print import Printer
from kirin.ir.attrs.abc import Attribute
from kirin.serialization.core.serializable import Serializable
from kirin.serialization.core.supportedtypes import SUPPORTED_PYTHON_TYPES
from kirin.serialization.core.serializationunit import SerializationUnit

if TYPE_CHECKING:
    from kirin.serialization.base.serializer import Serializer
    from kirin.serialization.base.deserializer import Deserializer

from .data import Data
from .types import PyClass, TypeAttribute

T = TypeVar("T")


@dataclass
class PyAttr(Data[T]):
    """Python attribute for compile-time values.
    This is a generic attribute that holds a Python value.

    The constructor takes a Python value and an optional type attribute.
    If the type attribute is not provided, the type of the value is inferred
    as `PyClass(type(value))`.

    !!! note "Pretty Printing"
        This object is pretty printable via
        [`.print()`][kirin.print.printable.Printable.print] method.
    """

    name = "PyAttr"
    data: T

    def __init__(self, data: T, pytype: TypeAttribute | None = None):
        self.data = data

        if pytype is None:
            self.type = PyClass(type(data))
        else:
            self.type = pytype

    def __hash__(self) -> int:
        return hash((self.type, self.data))

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, PyAttr):
            return False

        return self.type == value.type and self.data == value.data

    def print_impl(self, printer: Printer) -> None:
        printer.plain_print(repr(self.data))
        with printer.rich(style="comment"):
            printer.plain_print(" : ")
            printer.print(self.type)

    def unwrap(self) -> T:
        return self.data

    def is_structurally_equal(
        self, other: Attribute, context: dict | None = None
    ) -> bool:
        if not isinstance(other, PyAttr):
            return False
        if self.type != other.type:
            return False
        if isinstance(self.data, StructurallyEqual) and isinstance(
            other.data, StructurallyEqual
        ):

            return self.data.is_structurally_equal(other.data, context=context)
        return self.data == other.data

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        if not (
            isinstance(self.data, SUPPORTED_PYTHON_TYPES)
            or isinstance(self.data, Serializable)
        ):
            raise TypeError(
                f"Cannot serialize data of type {type(self.data)}. "
                "Data must be one of the supported Python types."
            )

        return SerializationUnit(
            kind="pyattr",
            module_name=self.__module__,
            class_name=self.__class__.__name__,
            data={
                "data": serializer.serialize(self.data),
                "pytype": serializer.serialize_attribute(self.type),
            },
        )

    @classmethod
    def deserialize(
        cls: Type["PyAttr"], serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> "PyAttr":
        pytype = deserializer.deserialize(serUnit.data["pytype"])
        value = deserializer.deserialize(serUnit.data["data"])
        return PyAttr(value, pytype=pytype)


@runtime_checkable
class StructurallyEqual(Protocol):
    def is_structurally_equal(
        self, other: Any, context: dict | None = None
    ) -> bool: ...



"""This module contains custom exception handling
for the Kirin-based compilers.
"""

from __future__ import annotations

import os
import sys
import math
import types
import shutil
import textwrap
from typing import TYPE_CHECKING

from rich.console import Console

from kirin.source import SourceInfo

if TYPE_CHECKING:
    from kirin import interp

KIRIN_INTERP_STATE = "__kirin_interp_state"
KIRIN_PYTHON_STACKTRACE = os.environ.get("KIRIN_PYTHON_STACKTRACE", "0") == "1"
KIRIN_STATIC_CHECK_LINENO = os.environ.get("KIRIN_STATIC_CHECK_LINENO", "1") == "1"
KIRIN_STATIC_CHECK_INDENT = int(os.environ.get("KIRIN_STATIC_CHECK_INDENT", "2"))
KIRIN_STATIC_CHECK_MAX_LINES = int(os.environ.get("KIRIN_STATIC_CHECK_MAX_LINES", "3"))


class StaticCheckError(Exception):
    def __init__(self, *messages: str, help: str | None = None) -> None:
        super().__init__(*messages)
        self.help: str | None = help
        self.source: SourceInfo | None = None
        self.lines: list[str] | None = None
        self.indent: int = KIRIN_STATIC_CHECK_INDENT
        self.max_lines: int = KIRIN_STATIC_CHECK_MAX_LINES
        self.show_lineno: bool = KIRIN_STATIC_CHECK_LINENO

    def hint(self):
        help = self.help or ""
        source = self.source or SourceInfo(0, 0, 0, 0)
        lines = self.lines or []
        begin = max(0, source.lineno - self.max_lines)
        end = max(
            max(source.lineno + self.max_lines, source.end_lineno or 1),
            1,
        )
        end = min(len(lines), end)  # make sure end is within bounds
        lines = lines[begin:end]
        error_lineno = source.lineno + source.lineno_begin
        error_lineno_len = len(str(error_lineno))
        code_indent = min(map(self.__get_indent, lines), default=0)

        console = Console(force_terminal=True)
        with console.capture() as capture:
            console.print(
                f"  {source or 'stdin'}",
                markup=True,
                highlight=False,
            )
            for lineno, line in enumerate(lines, begin):
                line = " " * self.indent + line[code_indent:]
                if self.show_lineno:
                    if lineno + 1 == source.lineno:
                        line = f"{error_lineno}[dim]│[/dim]" + line
                    else:
                        line = "[dim]" + " " * (error_lineno_len) + "│[/dim]" + line
                console.print("  " + line, markup=True, highlight=False)
                if lineno + 1 == source.lineno:
                    console.print(
                        "  "
                        + self.__arrow(
                            source,
                            code_indent,
                            error_lineno_len,
                            help,
                            self.indent,
                            self.show_lineno,
                        ),
                        markup=True,
                        highlight=False,
                    )
        return capture.get()

    def __arrow(
        self,
        source: SourceInfo,
        code_indent: int,
        error_lineno_len: int,
        help,
        indent: int,
        show_lineno: bool,
    ) -> str:
        ret = " " * (source.col_offset - code_indent)
        if source.end_col_offset:
            ret += "^" * (source.end_col_offset - source.col_offset)
        else:
            ret += "^"

        ret = " " * indent + "[red]" + ret
        if help:
            hint_indent = len(ret) - len("[ret]") + len(" help: ")
            terminal_width = math.floor(shutil.get_terminal_size().columns * 0.7)
            terminal_width = max(terminal_width - hint_indent, 10)
            wrapped = textwrap.fill(str(help), width=terminal_width)
            lines = wrapped.splitlines()
            ret += " help: " + lines[0] + "[/red]"
            for line in lines[1:]:
                ret += (
                    "\n"
                    + " " * (error_lineno_len + indent)
                    + "[dim]│[/dim]"
                    + " " * hint_indent
                    + "[red]"
                    + line
                    + "[/red]"
                )
        if show_lineno:
            ret = " " * error_lineno_len + "[dim]│[/dim]" + ret
        return ret

    @staticmethod
    def __get_indent(line: str) -> int:
        if len(line) == 0:
            return int(1e9)  # very large number
        return len(line) - len(line.lstrip())


def enable_stracetrace():
    """Enable the stacktrace for all exceptions."""
    global KIRIN_PYTHON_STACKTRACE
    KIRIN_PYTHON_STACKTRACE = True


def disable_stracetrace():
    """Disable the stacktrace for all exceptions."""
    global KIRIN_PYTHON_STACKTRACE
    KIRIN_PYTHON_STACKTRACE = False


def print_stacktrace(exception: Exception, state: interp.InterpreterState):
    frame: interp.FrameABC | None = state._current_frame
    print(
        "==== Python stacktrace has been disabled for simplicity, set KIRIN_PYTHON_STACKTRACE=1 to enable it ===="
    )
    print(f"{type(exception).__name__}: {exception}", file=sys.stderr)
    print("Traceback (most recent call last):", file=sys.stderr)
    frames: list[interp.FrameABC] = []
    while frame is not None:
        frames.append(frame)
        frame = frame.parent
    frames.reverse()
    for frame in frames:
        if stmt := frame.current_stmt:
            print("  " + repr(stmt.source), file=sys.stderr)
            print("     " + stmt.print_str(end=""), file=sys.stderr)


def exception_handler(exc_type, exc_value, exc_tb: types.TracebackType):
    """Custom exception handler to format and print exceptions."""
    if not KIRIN_PYTHON_STACKTRACE and issubclass(exc_type, StaticCheckError):
        console = Console(force_terminal=True)
        with console.capture() as capture:
            console.print(
                "==== Python stacktrace has been disabled for simplicity, set KIRIN_PYTHON_STACKTRACE=1 to enable it ===="
            )
            console.print(f"[bold red]{exc_type.__name__}:[/bold red]", end="")
        print(capture.get(), *exc_value.args, file=sys.stderr)
        print("Source Traceback:", file=sys.stderr)
        print(exc_value.hint(), file=sys.stderr, end="")
        return

    if (
        not KIRIN_PYTHON_STACKTRACE
        and (state := getattr(exc_value, KIRIN_INTERP_STATE, None)) is not None
    ):
        # Handle custom stack trace exceptions
        print_stacktrace(exc_value, state)
        return

    # Call the default exception handler
    sys.__excepthook__(exc_type, exc_value, exc_tb)


# Set the custom exception handler
sys.excepthook = exception_handler


def custom_exc(shell, etype, evalue, tb, tb_offset=None):
    if issubclass(etype, StaticCheckError):
        # Handle BuildError exceptions
        print(evalue, file=sys.stderr)
        return
    shell.showtraceback((etype, evalue, tb), tb_offset=tb_offset)


try:
    ip = get_ipython()  # type: ignore
    # Register your custom exception handler
    ip.set_custom_exc((Exception,), custom_exc)
except NameError:
    # Not in IPython, so we won't set the custom exception handler
    pass



# TODO: replace with something faster
from typing import Any, Generic, TypeVar, overload
from dataclasses import dataclass
from collections.abc import Sequence

from kirin import ir, types
from kirin.print.printer import Printer
from kirin.serialization.base.serializer import Serializer
from kirin.serialization.base.deserializer import Deserializer
from kirin.serialization.core.serializationunit import SerializationUnit

from ._dialect import dialect

T = TypeVar("T")
L = TypeVar("L")


@dataclass
@dialect.register
class IList(ir.Data[Sequence[T]], Sequence[T], Generic[T, L]):
    """A simple immutable list."""

    name = "IList"
    data: Sequence[T]
    elem: types.TypeAttribute = types.Any

    def __post_init__(self):
        self.type = types.Generic(IList, self.elem, types.Literal(len(self.data)))

    def __hash__(self) -> int:
        return id(self)  # do not hash the data

    def __len__(self) -> int:
        return len(self.data)

    @overload
    def __add__(self, other: "IList[T, Any]") -> "IList[T, Any]": ...

    @overload
    def __add__(self, other: list[T]) -> "IList[T, Any]": ...

    def __add__(self, other):
        if isinstance(other, list):
            return IList(list(self.data) + other, elem=self.elem)
        elif isinstance(other, IList):
            return IList(
                list(self.data) + list(other.data), elem=self.elem.join(other.elem)
            )
        else:
            raise TypeError(
                f"unsupported operand type(s) for +: 'IList' and '{type(other)}'"
            )

    @overload
    def __radd__(self, other: "IList[T, Any]") -> "IList[T, Any]": ...

    @overload
    def __radd__(self, other: list[T]) -> "IList[T, Any]": ...

    def __radd__(self, other):
        return IList(other + self.data)

    def __repr__(self) -> str:
        return f"IList({self.data})"

    def __str__(self) -> str:
        return f"IList({self.data})"

    def __iter__(self):
        return iter(self.data)

    @overload
    def __getitem__(self, index: int) -> T: ...

    @overload
    def __getitem__(self, index: slice) -> "IList[T, Any]": ...

    def __getitem__(self, index: int | slice) -> T | "IList[T, Any]":
        if isinstance(index, slice):
            return IList(self.data[index])
        return self.data[index]

    def __contains__(self, item: object) -> bool:
        return item in self.data

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, IList):
            return False
        return self.data == value.data

    def unwrap(self) -> Sequence[T]:
        return self

    def print_impl(self, printer: Printer) -> None:
        printer.print_seq(
            self.data, delim=", ", prefix="[", suffix="]", emit=printer.plain_print
        )
        printer.plain_print(")")

    def is_structurally_equal(
        self, other: ir.Attribute, context: dict | None = None
    ) -> bool:
        return (
            isinstance(other, IList)
            and self.data == other.data
            and self.elem.is_structurally_equal(other.elem, context=context)
        )

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return SerializationUnit(
            kind="ilist",
            module_name=dialect.name,
            class_name=IList.__name__,
            data={
                "data": serializer.serialize(self.data),
                "elem": serializer.serialize_attribute(self.elem),
            },
        )

    @classmethod
    def deserialize(
        cls, serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> "IList":
        items = deserializer.deserialize(serUnit.data["data"])
        elem = deserializer.deserialize(serUnit.data["elem"])
        return IList(items, elem=elem)



from kirin import lowering

from . import stmts as stmts, interp as interp
from ._dialect import dialect as dialect


@lowering.wraps(stmts.Random)
def random() -> float:
    """
    Generate a random floating number between 0 and 1.
    """
    ...


@lowering.wraps(stmts.RandInt)
def randint(start: int, stop: int) -> int:
    """
    Generate a random integer between the given range.
    """
    ...


@lowering.wraps(stmts.Uniform)
def uniform(start: float, stop: float) -> float:
    """
    Generate a random floating number between the given range.
    """
    ...


@lowering.wraps(stmts.Seed)
def seed(value: int) -> None:
    """
    Set the seed for the random number generator.
    """
    ...



# adopted from mkdocstrings/griffe-pydantic/
# from __future__ import annotations

from typing import Any, cast

import griffe

logger = griffe.get_logger(__name__)


def inherits_statement(cls: griffe.Class) -> bool:
    """Tell whether a class inherits from kirin.ir.Statement."""
    for base in cls.bases:
        if isinstance(base, (griffe.ExprName, griffe.Expr)):
            base = base.canonical_path  # noqa: PLW2901
        if base == "kirin.ir.Statement":
            return True
    return any(inherits_statement(parent_cls) for parent_cls in cls.mro())


def has_default(value: griffe.ExprCall) -> bool:
    return (
        len(value.arguments) >= 1
        and not isinstance(value.arguments[0], griffe.ExprKeyword)
        and value.arguments[0] != "..."
    )  # handle field(...), i.e. no default


def process_attribute(
    attr: griffe.Attribute, cls: griffe.Class, *, processed: set[str]
) -> None:
    """Handle ir.Statement attributes."""
    if attr.canonical_path in processed:
        return
    processed.add(attr.canonical_path)

    kwargs = {}
    if isinstance(attr.value, griffe.ExprCall):
        kwargs = {
            argument.name: argument.value
            for argument in attr.value.arguments
            if isinstance(argument, griffe.ExprKeyword)
        }
        field_descriptor = attr.value.function.canonical_path
        kwargs = _descriptor_args(attr.value)
        if field_descriptor == "kirin.decl.info.argument":
            attr.labels = {"kirin-argument"}
            if "kw_only" in kwargs and kwargs["kw_only"] == "True":
                attr.labels.add("kw-only")
        elif field_descriptor == "kirin.decl.info.attribute":
            if "property" in kwargs and kwargs["property"] == "True":
                attr.labels = {"kirin-property"}
            else:
                attr.labels = {"kirin-attribute"}
            attr.labels.add("kw-only")
        elif field_descriptor == "kirin.decl.info.result":
            attr.labels = {"kirin-result"}
        elif field_descriptor == "kirin.decl.info.region":
            attr.labels = {"kirin-region", "kw-only"}
        elif field_descriptor == "kirin.decl.info.block":
            attr.labels = {"kirin-block", "kw-only"}
    elif attr.value is not None:
        # logger.warning(f"Unhandled attribute value: {attr.value!r}")
        return  # wrong declaration
    elif attr.value is None and _is_annotation_SSAValue(attr.annotation):
        attr.labels = {"kirin-argument"}


def _is_annotation_SSAValue(annotation: str | griffe.Expr | None):
    if annotation is None:
        return False
    if isinstance(annotation, str):
        return annotation in ("ir.SSAValue", "SSAValue")
    if isinstance(annotation, griffe.Expr):
        return annotation.canonical_path in "kirin.ir.ssa.SSAValue"


def _descriptor_args(field: griffe.ExprCall) -> dict[str, Any]:
    kwargs = {
        argument.name: argument.value
        for argument in field.arguments
        if isinstance(argument, griffe.ExprKeyword)
    }
    if has_default(field):
        kwargs["type"] = field.arguments[0]
    return kwargs


def process_class(cls: griffe.Class, *, processed: set[str]) -> None:
    """Handle ir.Statement classes."""
    if cls.canonical_path in processed:
        return

    if not inherits_statement(cls):
        return

    processed.add(cls.canonical_path)
    cls.labels = {"krin-statement"}
    for member in cls.all_members.values():
        if isinstance(member, griffe.Attribute):
            process_attribute(member, cls, processed=processed)
        # TODO: process inherited methods?

    if "__init__" not in cls.members:
        _statement_init(cls)


def _statement_init_parameters(cls: griffe.Class) -> list[griffe.Parameter]:
    params: list[griffe.Parameter] = []
    for member in cls.members.values():
        if member.is_attribute:
            member = cast(griffe.Attribute, member)
            if member.annotation is None:
                continue

            if isinstance(member.value, griffe.ExprCall):
                kwargs = _descriptor_args(member.value)
            else:
                kwargs = {}

            if "init" in kwargs and kwargs["init"] == "False":
                continue

            kind = (
                griffe.ParameterKind.keyword_only
                if "kw-only" in member.labels
                else griffe.ParameterKind.positional_or_keyword
            )

            if "default_factory" in kwargs:
                default_factory = kwargs["default_factory"]
                if isinstance(default_factory, griffe.ExprLambda):
                    default = default_factory.body
                else:
                    default = griffe.ExprCall(
                        function=kwargs["default_factory"], arguments=[]
                    )
            else:
                default = None

            if "kirin-argument" in member.labels:
                params.append(
                    griffe.Parameter(
                        member.name,
                        annotation=member.annotation,
                        kind=kind,
                        default=default,
                        docstring=member.docstring,
                    )
                )
            elif (
                "kirin-attribute" in member.labels
                or "kirin-property" in member.labels
                or "kirin-block" in member.labels
                or "kirin-region" in member.labels
            ):
                params.append(
                    griffe.Parameter(
                        member.name,
                        annotation=member.annotation,
                        kind=griffe.ParameterKind.keyword_only,
                        default=default,
                        docstring=member.docstring,
                    )
                )
    return params


def _statement_init(cls: griffe.Class):
    params = []
    try:
        mro = cls.mro()
    except ValueError:
        mro = ()
    for parent in reversed(mro):
        if inherits_statement(parent):
            params.extend(_statement_init_parameters(parent))
            cls.labels.add("kirin-statement")

    if not inherits_statement(cls):
        return

    params.extend(_statement_init_parameters(cls))

    init = griffe.Function(
        "__init__",
        lineno=0,
        endlineno=0,
        parent=cls,
        parameters=griffe.Parameters(
            griffe.Parameter(
                name="self",
                annotation=None,
                kind=griffe.ParameterKind.positional_or_keyword,
                default=None,
                docstring=None,
            ),
            *_reorder_parameters(params),
        ),
        returns="None",
    )
    cls.set_member("__init__", init)


# NOTE: copied from griffe/src/_griffe/extensions/dataclasses.py
def _reorder_parameters(parameters: list[griffe.Parameter]) -> list[griffe.Parameter]:
    # De-duplicate, overwriting previous parameters.
    params_dict = {param.name: param for param in parameters}

    # Re-order, putting positional-only in front and keyword-only at the end.
    pos_only = []
    pos_kw = []
    kw_only = []
    for param in params_dict.values():
        if param.kind is griffe.ParameterKind.positional_only:
            pos_only.append(param)
        elif param.kind is griffe.ParameterKind.keyword_only:
            kw_only.append(param)
        else:
            pos_kw.append(param)
    return pos_only + pos_kw + kw_only


def process_module(mod: griffe.Module, *, processed: set[str]):
    if mod.canonical_path in processed:
        return

    processed.add(mod.canonical_path)
    for cls in mod.classes.values():
        if not cls.is_alias:
            process_class(cls, processed=processed)

    for submod in mod.modules.values():
        if not submod.is_alias:
            process_module(submod, processed=processed)


class KirinExtension(griffe.Extension):

    def __init__(self):
        super().__init__()
        self.in_statement: list[griffe.Class] = []
        self.processed: set[str] = set()

    def on_package_loaded(
        self, *, pkg: griffe.Module, loader: griffe.GriffeLoader, **kwargs: Any
    ) -> None:
        process_module(pkg, processed=self.processed)


# extensions = griffe.load_extensions(KirinExtension())
# data = griffe.load("kirin", extensions=extensions)

# obj = data["dialects.func.stmts.Function"]
# print(obj.labels)
# # print(obj.get_member("sym_name").labels)
# # print(obj.get_member("body").labels)
# # print(obj.get_member("sym_name").docstring.value)



from kirin.prelude import basic_no_opt, python_no_opt
from kirin.rewrite import Walk, Fixpoint, WrapConst
from kirin.analysis import const
from kirin.rewrite.fold import ConstantFold


@basic_no_opt
def foldable(x: int) -> int:
    y = 1
    b = y + 2
    c = y + b
    d = c + 4
    return d + x


def test_const_fold():
    before = foldable(1)
    const_prop = const.Propagate(foldable.dialects)
    frame, _ = const_prop.run(foldable)
    Fixpoint(Walk(WrapConst(frame))).rewrite(foldable.code)
    fold = ConstantFold()
    Fixpoint(Walk(fold)).rewrite(foldable.code)
    after = foldable(1)

    assert before == after


def test_const_fold_subroutine():

    @python_no_opt
    def non_pure_subroutine(x: list[int]) -> None:
        x.append(1)

    @python_no_opt
    def main():
        x = []
        non_pure_subroutine(x)
        x.append(2)

    old_main_region = main.callable_region.clone()

    fold = ConstantFold()
    Fixpoint(Walk(fold)).rewrite(main.code)

    assert old_main_region.is_structurally_equal(main.callable_region)



import pytest

from kirin.prelude import structural_no_opt
from kirin.testing import assert_structurally_same
from kirin.dialects import py, func


@pytest.fixture
def simple_example():
    """Kernel that computes 1*1 + 1."""

    @structural_no_opt
    def main():
        x = 1
        y = x * x
        z = y + x
        return z

    return main


def test_correct_list_passes_and_prints_nothing(simple_example, capfd):
    main = simple_example

    expected_statements = [
        x := py.Constant(1),
        y := py.Mult(x.result, x.result),
        z := py.Add(y.result, x.result),
        func.Return(z.result),
    ]
    assert_structurally_same(main, expected_statements)
    out, err = capfd.readouterr()
    assert out == ""
    assert err == ""


def test_debug_true_prints_something(simple_example, capfd):
    main = simple_example

    expected_statements = [
        x := py.Constant(1),
        y := py.Mult(x.result, x.result),
        z := py.Add(y.result, x.result),
        func.Return(z.result),
    ]
    assert_structurally_same(main, expected_statements, debug=True)
    out, err = capfd.readouterr()
    assert len(out.strip().split("\n")) == 4
    assert err == ""


def test_wrong_length_fails(simple_example):
    main = simple_example

    expected_statements = [
        x := py.Constant(1),
        y := py.Mult(x.result, x.result),
        py.Add(y.result, x.result),
    ]
    with pytest.raises(AssertionError):
        assert_structurally_same(main, expected_statements)


def test_wrong_argument_fails_by_default(simple_example):
    main = simple_example

    expected_statements = [
        x := py.Constant(1),
        y := py.Mult(x.result, x.result),
        z := py.Add(y.result, y.result),  # Wrong second argument.
        func.Return(z.result),
    ]
    with pytest.raises(AssertionError):
        assert_structurally_same(main, expected_statements)


def test_wrong_argument_passes_if_check_args_false(simple_example):
    main = simple_example

    expected_statements = [
        x := py.Constant(1),
        y := py.Mult(x.result, x.result),
        z := py.Add(y.result, y.result),  # Wrong second argument.
        func.Return(z.result),
    ]
    assert_structurally_same(main, expected_statements, check_args=False)


def test_wrong_statement_raises_assertion_error(simple_example):
    main = simple_example

    expected_statements = [
        x := py.Constant(1),
        y := py.Mult(x.result, x.result),
        z := py.Mult(y.result, x.result),  # Wrong statement.
        func.Return(z.result),
    ]
    with pytest.raises(AssertionError):
        assert_structurally_same(main, expected_statements)



"""Boolean operators for Python dialect.

This module contains the dialect for the Python boolean operators, including:

- The `And` and `Or` statement classes.
- The lowering pass for the boolean operators.
- The concrete implementation of the boolean operators.
- The Julia emitter for the boolean operators.

This dialect maps `ast.BoolOp` nodes to the `And` and `Or` statements.
"""

import ast

from kirin import ir, types, interp, lowering
from kirin.decl import info, statement

dialect = ir.Dialect("py.boolop")


@statement
class BoolOp(ir.Statement):
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    lhs: ir.SSAValue = info.argument(print=False)
    rhs: ir.SSAValue = info.argument(print=False)
    result: ir.ResultValue = info.result(types.Bool)


@statement(dialect=dialect)
class And(BoolOp):
    name = "and"


@statement(dialect=dialect)
class Or(BoolOp):
    name = "or"


@dialect.register
class PythonLowering(lowering.FromPythonAST):

    def lower_BoolOp(self, state: lowering.State, node: ast.BoolOp) -> lowering.Result:
        lhs = state.lower(node.values[0]).expect_one()
        match node.op:
            case ast.And():
                boolop = And
            case ast.Or():
                boolop = Or
            case _:
                raise lowering.BuildError(f"unsupported boolop {node.op}")

        for value in node.values[1:]:
            lhs = state.current_frame.push(
                boolop(lhs=lhs, rhs=state.lower(value).expect_one())
            ).result
        return lhs


@dialect.register
class BoolOpMethod(interp.MethodTable):

    @interp.impl(And)
    def and_(self, interp, frame: interp.Frame, stmt: And):
        return (frame.get(stmt.lhs) and frame.get(stmt.rhs),)

    @interp.impl(Or)
    def or_(self, interp, frame: interp.Frame, stmt: Or):
        return (frame.get(stmt.lhs) or frame.get(stmt.rhs),)



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar, cast
from dataclasses import field, dataclass

from typing_extensions import Self, dataclass_transform

from kirin.ir.nodes import Statement
from kirin.ir.attrs.abc import Attribute

T = TypeVar("T")

if TYPE_CHECKING:
    from kirin.types import PyClass
    from kirin.rewrite.abc import RewriteRule
    from kirin.interp.table import MethodTable
    from kirin.lowering.python.dialect import FromPythonAST
    from kirin.serialization.base.serializer import Serializer
    from kirin.serialization.base.deserializer import Deserializer
    from kirin.serialization.core.serializationunit import SerializationUnit


@dataclass
class Rules:
    canonicalize: list[RewriteRule] = field(default_factory=list, init=True)
    """A collection of rules for Canonicalize pass."""
    inference: list[RewriteRule] = field(default_factory=list, init=True)
    """A collection of rules for Inference pass."""


# TODO: add an option to generate default lowering at dialect construction
@dataclass
class Dialect:
    """Dialect is a collection of statements, attributes, interpreters, lowerings, and codegen.

    Example:
        ```python
            from kirin import ir

            my_dialect = ir.Dialect(name="my_dialect")

        ```
    """

    name: str
    """The name of the dialect."""
    stmts: list[type[Statement]] = field(default_factory=list, init=True)
    """A list of statements in the dialect."""
    attrs: list[type[Attribute]] = field(default_factory=list, init=True)
    """A list of attributes in the dialect."""
    interps: dict[str, MethodTable] = field(default_factory=dict, init=True)
    """A dictionary of registered method table in the dialect."""
    lowering: dict[str, FromPythonAST] = field(default_factory=dict, init=True)
    """A dictionary of registered python lowering implmentations in the dialect."""
    rules: Rules = field(default_factory=Rules, init=True)
    """A collection of rewrite rules for the dialect."""
    python_types: dict[tuple[str, str], "PyClass"] = field(
        default_factory=dict, init=True
    )

    def __post_init__(self) -> None:
        from kirin.lowering.python.dialect import NoSpecialLowering

        self.lowering["default"] = NoSpecialLowering()

    def __repr__(self) -> str:
        return f"Dialect(name={self.name}, ...)"

    def __hash__(self) -> int:
        return hash(self.name)

    @dataclass_transform()
    def register(self, node: type | None = None, key: str | None = None):
        """register is a decorator to register a node to the dialect.

        Args:
            node (type | None): The node to register. Defaults to None.
            key (str | None): The key to register the node to. Defaults to None.

        Raises:
            ValueError: If the node is not a subclass of Statement, Attribute, DialectInterpreter, FromPythonAST, or DialectEmit.

        Example:
            * Register a method table for concrete interpreter (by default key="main") to the dialect:
            ```python
                from kirin import ir

                my_dialect = ir.Dialect(name="my_dialect")

                @my_dialect.register
                class MyMethodTable(ir.MethodTable):
                    ...
            ```

            * Register a method table for the interpreter specified by `key` to the dialect:
            ```python
                from kirin import ir

                my_dialect = ir.Dialect(name="my_dialect")

                @my_dialect.register(key="my_interp")
                class MyMethodTable(ir.MethodTable):
                    ...
            ```


        """
        from kirin.interp.table import MethodTable
        from kirin.lowering.python.dialect import FromPythonAST

        if key is None:
            key = "main"

        def wrapper(node: type[T]) -> type[T]:
            if issubclass(node, Statement):
                self.stmts.append(node)
            elif issubclass(node, Attribute):
                assert (
                    Attribute in node.__mro__
                ), f"{node} is not a subclass of Attribute"
                setattr(node, "dialect", self)
                assert hasattr(node, "name"), f"{node} does not have a name attribute"
                self.attrs.append(node)
            elif issubclass(node, MethodTable):
                if key in self.interps:
                    raise ValueError(
                        f"Cannot register {node} to Dialect, key {key} exists in {self}"
                    )
                self.interps[key] = node()
            elif issubclass(node, FromPythonAST):
                if key in self.lowering:
                    raise ValueError(
                        f"Cannot register {node} to Dialect, key {key} exists"
                    )
                self.lowering[key] = node()
            else:
                raise ValueError(f"Cannot register {node} to Dialect")
            return node

        if node is None:
            return wrapper

        return wrapper(node)

    def register_py_type(
        self,
        node: type[T] | "PyClass[T]",
        display_name: str | None = None,
        prefix: str = "py",
    ):
        from kirin.ir.attrs.types import PyClass

        if isinstance(node, type):
            node = PyClass(node, display_name=display_name, prefix=prefix)

        if isinstance(node, PyClass):
            if (node.prefix, node.display_name) in self.python_types and (
                other_node := self.python_types[(node.prefix, node.display_name)]
            ) != node:
                raise ValueError(
                    f"Cannot register {node} to Dialect, type {other_node.prefix}.{other_node.display_name} exists for {other_node.typ}"
                )

            self.python_types[(node.prefix, node.display_name)] = node
            return node

        else:
            raise ValueError(f"Cannot register {node} to Dialect")

    def canonicalize(self, rule: type[RewriteRule]) -> type[RewriteRule]:
        """Register a rewrite rule to the canonicalization pass.

        Args:
            rule (RewriteRule): The rewrite rule to register.
        """
        self.rules.canonicalize.append(rule())
        return rule

    def post_inference(self, rule: type[RewriteRule]) -> type[RewriteRule]:
        """Register a rewrite rule to the inference pass.
        Usually, this is used to register a rule that requires
        type inference to be run first.

        Args:
            rule (RewriteRule): The rewrite rule to register.
        """
        self.rules.inference.append(rule())
        return rule

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return serializer.serialize_dialect(self)

    @classmethod
    def deserialize(
        cls: type[Self], serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> Self:
        return cast(Self, deserializer.deserialize_dialect(serUnit))



from abc import abstractmethod
from typing import TYPE_CHECKING, Protocol, runtime_checkable

from typing_extensions import Self

if TYPE_CHECKING:
    from kirin.serialization.base.deserializer import Deserializer
    from kirin.serialization.core.serializationunit import SerializationUnit


@runtime_checkable
class Deserializable(Protocol):

    @classmethod
    @abstractmethod
    def deserialize(
        cls: type[Self], serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> Self:
        raise NotImplementedError



from kirin.prelude import basic
from kirin.dialects import ilist
from kirin.serialization.jsonserializer import JSONSerializer


@basic
def foo(x: int, y: float, z: bool):
    c = [[(200.0, 200.0), (210.0, 200.0)]]
    if z:
        c = [(222.0, 333.0)]
    else:
        return [1, 2, 3, 4]
    return c


@basic
def bar():
    def goo(x: int):
        a = (3, 4)
        return a[0]

    def boo(y):
        return goo(y) + 1

    boo(4)


@basic
def loop_ilist():
    a = 0
    c = ilist.IList([a, a * 2])
    for i in range(3):
        a = i
        c = ilist.IList([a, a * 2])
    return c


@basic
def my_kernel1(x: int):
    return (x, x + 1, 3)


@basic
def my_kernel2(y: int):
    return my_kernel1(y) * 10


@basic
def foo2(y: int):

    def inner(x: int):
        return x * y + 1

    return inner


inner_ker = foo2(y=10)


@basic
def main_lambda(z: int):
    return inner_ker(z)


@basic
def slicing():
    in1 = ("a", "b", "c", "d", "e", "f", "g", "h")
    in2 = [1, 2, 3, 4, 5]

    x = slice(3, 5)
    a = in2[x]
    b = in1[1:4]
    c = in1[:3]
    d = in1[2:]
    e = in1[:]
    return (a, b, c, d, e)


def test_round_trip_sequence_run():
    encoded = basic.encode(slicing)
    decoded = basic.decode(encoded)
    before = slicing()
    after = decoded()
    assert before == after

    json_ser = JSONSerializer()
    json_encoded = json_ser.encode(encoded)
    json_decoded = json_ser.decode(json_encoded)
    decoded_2 = basic.decode(json_decoded)
    after2 = decoded_2()
    assert before == after2 == after


def test_round_trip1_run():
    encoded = basic.encode(my_kernel1)
    decoded = basic.decode(encoded)
    before = my_kernel1(10)
    after = decoded(10)
    assert before == after
    json_ser = JSONSerializer()
    json_encoded = json_ser.encode(encoded)
    json_decoded = json_ser.decode(json_encoded)
    decoded_2 = basic.decode(json_decoded)
    after2 = decoded_2(10)
    assert before == after2 == after


def test_round_trip2_run():
    encoded = basic.encode(foo)
    decoded = basic.decode(encoded)
    before = foo(10, 20.0, True)
    after = decoded(10, 20.0, True)
    assert before == after
    json_ser = JSONSerializer()
    json_encoded = json_ser.encode(encoded)
    json_decoded = json_ser.decode(json_encoded)
    decoded_2 = basic.decode(json_decoded)
    after2 = decoded_2(10, 20.0, True)
    assert before == after2 == after


def test_round_trip3_run():
    encoded = basic.encode(main_lambda)
    decoded = basic.decode(encoded)
    before = main_lambda(2)
    after = decoded(2)
    assert before == after

    json_ser = JSONSerializer()
    json_encoded = json_ser.encode(encoded)
    json_decoded = json_ser.decode(json_encoded)
    decoded_2 = basic.decode(json_decoded)
    after2 = decoded_2(2)
    assert before == after2 == after



from .serializable import Serializable as Serializable
from .deserializable import Deserializable as Deserializable
from .serializationunit import SerializationUnit as SerializationUnit
from .serializationmodule import SerializationModule as SerializationModule



"""SSACFG region trait.

This module defines the SSACFGRegion trait, which is used to indicate that a
region has an SSACFG graph.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar
from dataclasses import dataclass

from kirin.ir.ssa import SSAValue as SSAValue
from kirin.ir.traits.abc import RegionGraph, RegionInterpretationTrait
from kirin.ir.nodes.region import Region

if TYPE_CHECKING:
    from kirin import ir
    from kirin.interp.frame import FrameABC


@dataclass(frozen=True)
class HasCFG(RegionGraph):

    def get_graph(self, region: ir.Region):
        from kirin.analysis.cfg import CFG

        return CFG(region)


@dataclass(frozen=True)
class SSACFG(RegionInterpretationTrait):

    ValueType = TypeVar("ValueType")

    @classmethod
    def set_region_input(
        cls, frame: FrameABC[SSAValue, ValueType], region: Region, *inputs: ValueType
    ) -> None:
        frame.set_values(region.blocks[0].args, inputs)



from dataclasses import dataclass

from kirin import ir
from kirin.analysis import const
from kirin.rewrite.abc import RewriteRule, RewriteResult
from kirin.dialects.func import Call, Invoke


@dataclass
class Call2Invoke(RewriteRule):
    """Rewrite a `Call` statement to an `Invoke` statement."""

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not isinstance(node, Call):
            return RewriteResult()

        if (mt := node.callee.hints.get("const")) is None:
            return RewriteResult()

        if not isinstance(mt, const.Value):
            return RewriteResult()

        if not isinstance(mt.data, ir.Method):
            return RewriteResult()

        method = mt.data
        trait = method.code.get_present_trait(ir.CallableStmtInterface)
        inputs = trait.align_input_args(
            method.code, *node.inputs, **dict(zip(node.keys, node.kwargs))
        )
        stmt = Invoke(inputs=inputs, callee=mt.data)
        for result, new_result in zip(node.results, stmt.results):
            new_result.name = result.name
            new_result.type = result.type
            if result_hint := result.hints.get("const"):
                new_result.hints["const"] = result_hint

        node.replace_by(stmt)
        return RewriteResult(has_done_something=True)



# This file is generated by gen.py
from kirin import ir, types, lowering
from kirin.decl import info, statement
from kirin.dialects.math.dialect import dialect


@statement(dialect=dialect)
class acos(ir.Statement):
    """acos statement, wrapping the math.acos function"""

    name = "acos"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class asin(ir.Statement):
    """asin statement, wrapping the math.asin function"""

    name = "asin"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class asinh(ir.Statement):
    """asinh statement, wrapping the math.asinh function"""

    name = "asinh"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class atan(ir.Statement):
    """atan statement, wrapping the math.atan function"""

    name = "atan"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class atan2(ir.Statement):
    """atan2 statement, wrapping the math.atan2 function"""

    name = "atan2"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    y: ir.SSAValue = info.argument(types.Float)
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class atanh(ir.Statement):
    """atanh statement, wrapping the math.atanh function"""

    name = "atanh"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class ceil(ir.Statement):
    """ceil statement, wrapping the math.ceil function"""

    name = "ceil"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class copysign(ir.Statement):
    """copysign statement, wrapping the math.copysign function"""

    name = "copysign"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    y: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class cos(ir.Statement):
    """cos statement, wrapping the math.cos function"""

    name = "cos"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class cosh(ir.Statement):
    """cosh statement, wrapping the math.cosh function"""

    name = "cosh"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class degrees(ir.Statement):
    """degrees statement, wrapping the math.degrees function"""

    name = "degrees"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class erf(ir.Statement):
    """erf statement, wrapping the math.erf function"""

    name = "erf"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class erfc(ir.Statement):
    """erfc statement, wrapping the math.erfc function"""

    name = "erfc"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class exp(ir.Statement):
    """exp statement, wrapping the math.exp function"""

    name = "exp"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class expm1(ir.Statement):
    """expm1 statement, wrapping the math.expm1 function"""

    name = "expm1"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class fabs(ir.Statement):
    """fabs statement, wrapping the math.fabs function"""

    name = "fabs"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class floor(ir.Statement):
    """floor statement, wrapping the math.floor function"""

    name = "floor"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class fmod(ir.Statement):
    """fmod statement, wrapping the math.fmod function"""

    name = "fmod"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    y: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class gamma(ir.Statement):
    """gamma statement, wrapping the math.gamma function"""

    name = "gamma"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class isfinite(ir.Statement):
    """isfinite statement, wrapping the math.isfinite function"""

    name = "isfinite"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Bool)


@statement(dialect=dialect)
class isinf(ir.Statement):
    """isinf statement, wrapping the math.isinf function"""

    name = "isinf"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Bool)


@statement(dialect=dialect)
class isnan(ir.Statement):
    """isnan statement, wrapping the math.isnan function"""

    name = "isnan"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Bool)


@statement(dialect=dialect)
class lgamma(ir.Statement):
    """lgamma statement, wrapping the math.lgamma function"""

    name = "lgamma"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class log10(ir.Statement):
    """log10 statement, wrapping the math.log10 function"""

    name = "log10"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class log1p(ir.Statement):
    """log1p statement, wrapping the math.log1p function"""

    name = "log1p"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class log2(ir.Statement):
    """log2 statement, wrapping the math.log2 function"""

    name = "log2"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class pow(ir.Statement):
    """pow statement, wrapping the math.pow function"""

    name = "pow"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    y: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class radians(ir.Statement):
    """radians statement, wrapping the math.radians function"""

    name = "radians"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class remainder(ir.Statement):
    """remainder statement, wrapping the math.remainder function"""

    name = "remainder"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    y: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class sin(ir.Statement):
    """sin statement, wrapping the math.sin function"""

    name = "sin"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class sinh(ir.Statement):
    """sinh statement, wrapping the math.sinh function"""

    name = "sinh"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class sqrt(ir.Statement):
    """sqrt statement, wrapping the math.sqrt function"""

    name = "sqrt"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class tan(ir.Statement):
    """tan statement, wrapping the math.tan function"""

    name = "tan"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class tanh(ir.Statement):
    """tanh statement, wrapping the math.tanh function"""

    name = "tanh"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class trunc(ir.Statement):
    """trunc statement, wrapping the math.trunc function"""

    name = "trunc"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)


@statement(dialect=dialect)
class ulp(ir.Statement):
    """ulp statement, wrapping the math.ulp function"""

    name = "ulp"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    x: ir.SSAValue = info.argument(types.Float)
    result: ir.ResultValue = info.result(types.Float)



from pytest import mark

from kirin import types
from kirin.prelude import basic, structural
from kirin.dialects import ilist


@mark.xfail(reason="if with early return not supported in scf lowering")
def test_inter_method_infer():
    @basic
    def foo(x: int):
        if x > 1:
            return x + 1
        else:
            return x - 1.0

    @basic(typeinfer=True, no_raise=False)
    def main(x: int):
        return foo(x)

    @basic(typeinfer=True, no_raise=False)
    def moo(x):
        return foo(x)

    assert main.return_type == (types.Int | types.Float)
    # assert moo.arg_types[0] == types.Int  # type gets narrowed based on callee
    assert moo.return_type == (types.Int | types.Float)
    # NOTE: inference of moo should not update foo
    assert foo.arg_types[0] == types.Int
    assert foo.inferred is False
    assert foo.return_type is types.Any


@mark.xfail(reason="if with early return not supported in scf lowering")
def test_infer_if_return():
    from kirin.prelude import structural

    @structural(typeinfer=True, fold=True, no_raise=False)
    def test(b: bool):
        if b:
            return False
        else:
            b = not b

        return b

    test.print()


def test_method_constant_type_infer():

    @structural(typeinfer=True, fold=False)
    def _new(qid: int):
        return 1

    @structural(fold=False, typeinfer=True)
    def alloc(n_iter: int):
        return ilist.map(_new, ilist.range(n_iter))

    assert alloc.return_type.is_subseteq(ilist.IListType[types.Literal(1), types.Any])



from kirin import types
from kirin.prelude import basic

x = [1, 2, 3]


@basic(typeinfer=True)
def main():
    return x[1]


main.print(hint="const")


def test_const_infer():
    assert main.return_type is not None
    assert main.return_type.is_subseteq(types.Int)



from kirin import ir, types
from kirin.prelude import structural
from kirin.dialects import ilist


def test_infer_lambda():
    @structural(typeinfer=True, fold=False, no_raise=False)
    def main(n):
        def map_func(i):
            return n + 1

        return ilist.map(map_func, ilist.range(4))

    map_stmt = main.callable_region.blocks[0].stmts.at(-2)
    assert isinstance(map_stmt, ilist.Map)
    assert map_stmt.result.type == ilist.IListType[types.Int, types.Literal(4)]


def test_infer_method_type_hint_call():

    @structural(typeinfer=True, fold=False, no_raise=False)
    def main(n, fx: ir.Method[[int], int]):
        return fx(n)

    assert main.return_type == types.Int


def test_infer_method_type_hint():

    @structural(typeinfer=True, fold=False, no_raise=False)
    def main(n, fx: ir.Method[[int], int]):
        def map_func(i):
            return n + 1 + fx(i)

        return ilist.map(map_func, ilist.range(4))

    assert main.return_type == ilist.IListType[types.Int, types.Literal(4)]



from typing import TYPE_CHECKING, Generic, TypeVar
from dataclasses import dataclass

from kirin import types
from kirin.ir import Attribute
from kirin.print.printer import Printer
from kirin.serialization.core.serializationunit import SerializationUnit

if TYPE_CHECKING:
    from kirin.serialization.base.serializer import Serializer
    from kirin.serialization.base.deserializer import Deserializer

from ._dialect import dialect

TypeLatticeElem = TypeVar("TypeLatticeElem", bound="types.TypeAttribute")


@dialect.register
@dataclass
class Signature(Generic[TypeLatticeElem], Attribute):
    """function body signature.

    This is not a type attribute because it just stores
    the signature of a function at its definition site.
    We don't perform type inference on this directly.

    The type of a function is the type of `inputs[0]`, which
    typically is a `MethodType`.
    """

    name = "Signature"
    inputs: tuple[TypeLatticeElem, ...]
    output: TypeLatticeElem  # multi-output must be tuple

    def __hash__(self) -> int:
        return hash((self.inputs, self.output))

    def print_impl(self, printer: Printer) -> None:
        printer.print_seq(self.inputs, delim=", ", prefix="(", suffix=")")
        printer.plain_print(" -> ")
        printer.print(self.output)

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, Signature):
            return False
        return self.inputs == value.inputs and self.output == value.output

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return SerializationUnit(
            kind="signature",
            module_name=self.__module__,
            class_name=self.__class__.__name__,
            data={
                "inputs": serializer.serialize_tuple(self.inputs),
                "output": serializer.serialize(self.output),
            },
        )

    def is_structurally_equal(
        self,
        other: Attribute,
        context: dict | None = None,
    ) -> bool:
        return self == other

    @classmethod
    def deserialize(
        cls, serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> "Signature":
        inputs = deserializer.deserialize(serUnit.data["inputs"])
        output = deserializer.deserialize(serUnit.data["output"])
        return Signature(inputs=inputs, output=output)



from typing import Union
from collections.abc import Sequence

SUPPORTED_PYTHON_TYPES = Union[
    # Python built-in types
    bool,
    bytes,
    bytearray,
    dict,
    float,
    frozenset,
    int,
    list,
    range,
    set,
    slice,
    str,
    tuple,
    type,
    type(None),
    Sequence,
]
